# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Evaluation dataset creation functions."""

import csv
import functools
import io
import os
from os import path
import pickle
import random
from typing import Iterable, Mapping, Optional, Tuple, Union

from absl import logging

from kubric.challenges.point_tracking import dataset
import mediapy as media
import numpy as np
from PIL import Image
import scipy.io as sio
import tensorflow as tf
import tensorflow_datasets as tfds

from juxtapose.trackers.tapnet.utils import transforms

DatasetElement = Mapping[str, Mapping[str, Union[np.ndarray, str]]]


def resize_video(video: np.ndarray, output_size: Tuple[int, int]) -> np.ndarray:
    """Resize a video to output_size."""
    # If you have a GPU, consider replacing this with a GPU-enabled resize op,
    # such as a jitted jax.image.resize.  It will make things faster.
    return media.resize_video(video, output_size[1:3])


def compute_tapvid_metrics(
    query_points: np.ndarray,
    gt_occluded: np.ndarray,
    gt_tracks: np.ndarray,
    pred_occluded: np.ndarray,
    pred_tracks: np.ndarray,
    query_mode: str,
) -> Mapping[str, np.ndarray]:
    """Computes TAP-Vid metrics (Jaccard, Pts.

    Within Thresh, Occ.

    Acc.)

    See the TAP-Vid paper for details on the metric computation.  All inputs are
    given in raster coordinates.  The first three arguments should be the direct
    outputs of the reader: the 'query_points', 'occluded', and 'target_points'.
    The paper metrics assume these are scaled relative to 256x256 images.
    pred_occluded and pred_tracks are your algorithm's predictions.

    This function takes a batch of inputs, and computes metrics separately for
    each video.  The metrics for the full benchmark are a simple mean of the
    metrics across the full set of videos.  These numbers are between 0 and 1,
    but the paper multiplies them by 100 to ease reading.

    Args:
       query_points: The query points, an in the format [t, y, x].  Its size is
         [b, n, 3], where b is the batch size and n is the number of queries
       gt_occluded: A boolean array of shape [b, n, t], where t is the number of
         frames.  True indicates that the point is occluded.
       gt_tracks: The target points, of shape [b, n, t, 2].  Each point is in the
         format [x, y]
       pred_occluded: A boolean array of predicted occlusions, in the same format
         as gt_occluded.
       pred_tracks: An array of track predictions from your algorithm, in the same
         format as gt_tracks.
       query_mode: Either 'first' or 'strided', depending on how queries are
         sampled.  If 'first', we assume the prior knowledge that all points
         before the query point are occluded, and these are removed from the
         evaluation.

    Returns:
        A dict with the following keys:

        occlusion_accuracy: Accuracy at predicting occlusion.
        pts_within_{x} for x in [1, 2, 4, 8, 16]: Fraction of points
          predicted to be within the given pixel threshold, ignoring occlusion
          prediction.
        jaccard_{x} for x in [1, 2, 4, 8, 16]: Jaccard metric for the given
          threshold
        average_pts_within_thresh: average across pts_within_{x}
        average_jaccard: average across jaccard_{x}
    """

    metrics = {}

    eye = np.eye(gt_tracks.shape[2], dtype=np.int32)
    if query_mode == "first":
        # evaluate frames after the query frame
        query_frame_to_eval_frames = np.cumsum(eye, axis=1) - eye
    elif query_mode == "strided":
        # evaluate all frames except the query frame
        query_frame_to_eval_frames = 1 - eye
    else:
        raise ValueError("Unknown query mode " + query_mode)

    query_frame = query_points[..., 0]
    query_frame = np.round(query_frame).astype(np.int32)
    evaluation_points = query_frame_to_eval_frames[query_frame] > 0

    # Occlusion accuracy is simply how often the predicted occlusion equals the
    # ground truth.
    occ_acc = np.sum(
        np.equal(pred_occluded, gt_occluded) & evaluation_points,
        axis=(1, 2),
    ) / np.sum(evaluation_points)
    metrics["occlusion_accuracy"] = occ_acc

    # Next, convert the predictions and ground truth positions into pixel
    # coordinates.
    visible = np.logical_not(gt_occluded)
    pred_visible = np.logical_not(pred_occluded)
    all_frac_within = []
    all_jaccard = []
    for thresh in [1, 2, 4, 8, 16]:
        # True positives are points that are within the threshold and where both
        # the prediction and the ground truth are listed as visible.
        within_dist = np.sum(
            np.square(pred_tracks - gt_tracks),
            axis=-1,
        ) < np.square(thresh)
        is_correct = np.logical_and(within_dist, visible)

        # Compute the frac_within_threshold, which is the fraction of points
        # within the threshold among points that are visible in the ground truth,
        # ignoring whether they're predicted to be visible.
        count_correct = np.sum(
            is_correct & evaluation_points,
            axis=(1, 2),
        )
        count_visible_points = np.sum(visible & evaluation_points, axis=(1, 2))
        frac_correct = count_correct / count_visible_points
        metrics["pts_within_" + str(thresh)] = frac_correct
        all_frac_within.append(frac_correct)

        true_positives = np.sum(
            is_correct & pred_visible & evaluation_points, axis=(1, 2)
        )

        # The denominator of the jaccard metric is the true positives plus
        # false positives plus false negatives.  However, note that true positives
        # plus false negatives is simply the number of points in the ground truth
        # which is easier to compute than trying to compute all three quantities.
        # Thus we just add the number of points in the ground truth to the number
        # of false positives.
        #
        # False positives are simply points that are predicted to be visible,
        # but the ground truth is not visible or too far from the prediction.
        gt_positives = np.sum(visible & evaluation_points, axis=(1, 2))
        false_positives = (~visible) & pred_visible
        false_positives = false_positives | ((~within_dist) & pred_visible)
        false_positives = np.sum(false_positives & evaluation_points, axis=(1, 2))
        jaccard = true_positives / (gt_positives + false_positives)
        metrics["jaccard_" + str(thresh)] = jaccard
        all_jaccard.append(jaccard)
    metrics["average_jaccard"] = np.mean(
        np.stack(all_jaccard, axis=1),
        axis=1,
    )
    metrics["average_pts_within_thresh"] = np.mean(
        np.stack(all_frac_within, axis=1),
        axis=1,
    )
    return metrics


def latex_table(mean_scalars: Mapping[str, float]) -> str:
    """Generate a latex table for displaying TAP-Vid and PCK metrics."""
    if "average_jaccard" in mean_scalars:
        latex_fields = [
            "average_jaccard",
            "average_pts_within_thresh",
            "occlusion_accuracy",
            "jaccard_1",
            "jaccard_2",
            "jaccard_4",
            "jaccard_8",
            "jaccard_16",
            "pts_within_1",
            "pts_within_2",
            "pts_within_4",
            "pts_within_8",
            "pts_within_16",
        ]
        header = (
            "AJ & $<\\delta^{x}_{avg}$ & OA & Jac. $\\delta^{0}$ & "
            + "Jac. $\\delta^{1}$ & Jac. $\\delta^{2}$ & "
            + "Jac. $\\delta^{3}$ & Jac. $\\delta^{4}$ & $<\\delta^{0}$ & "
            + "$<\\delta^{1}$ & $<\\delta^{2}$ & $<\\delta^{3}$ & "
            + "$<\\delta^{4}$"
        )
    else:
        latex_fields = ["PCK@0.1", "PCK@0.2", "PCK@0.3", "PCK@0.4", "PCK@0.5"]
        header = " & ".join(latex_fields)

    body = " & ".join(
        [f"{float(np.array(mean_scalars[x]*100)):.3}" for x in latex_fields]
    )
    return "\n".join([header, body])


def sample_queries_strided(
    target_occluded: np.ndarray,
    target_points: np.ndarray,
    frames: np.ndarray,
    query_stride: int = 5,
) -> Mapping[str, np.ndarray]:
    """Package a set of frames and tracks for use in TAPNet evaluations.

    Given a set of frames and tracks with no query points, sample queries
    strided every query_stride frames, ignoring points that are not visible
    at the selected frames.

    Args:
      target_occluded: Boolean occlusion flag, of shape [n_tracks, n_frames],
        where True indicates occluded.
      target_points: Position, of shape [n_tracks, n_frames, 2], where each point
        is [x,y] scaled between 0 and 1.
      frames: Video tensor, of shape [n_frames, height, width, 3].  Scaled between
        -1 and 1.
      query_stride: When sampling query points, search for un-occluded points
        every query_stride frames and convert each one into a query.

    Returns:
      A dict with the keys:
        video: Video tensor of shape [1, n_frames, height, width, 3].  The video
          has floats scaled to the range [-1, 1].
        query_points: Query points of shape [1, n_queries, 3] where
          each point is [t, y, x] scaled to the range [-1, 1].
        target_points: Target points of shape [1, n_queries, n_frames, 2] where
          each point is [x, y] scaled to the range [-1, 1].
        trackgroup: Index of the original track that each query point was
          sampled from.  This is useful for visualization.
    """
    tracks = []
    occs = []
    queries = []
    trackgroups = []
    total = 0
    trackgroup = np.arange(target_occluded.shape[0])
    for i in range(0, target_occluded.shape[1], query_stride):
        mask = target_occluded[:, i] == 0
        query = np.stack(
            [
                i * np.ones(target_occluded.shape[0:1]),
                target_points[:, i, 1],
                target_points[:, i, 0],
            ],
            axis=-1,
        )
        queries.append(query[mask])
        tracks.append(target_points[mask])
        occs.append(target_occluded[mask])
        trackgroups.append(trackgroup[mask])
        total += np.array(np.sum(target_occluded[:, i] == 0))

    return {
        "video": frames[np.newaxis, ...],
        "query_points": np.concatenate(queries, axis=0)[np.newaxis, ...],
        "target_points": np.concatenate(tracks, axis=0)[np.newaxis, ...],
        "occluded": np.concatenate(occs, axis=0)[np.newaxis, ...],
        "trackgroup": np.concatenate(trackgroups, axis=0)[np.newaxis, ...],
    }


def sample_queries_first(
    target_occluded: np.ndarray,
    target_points: np.ndarray,
    frames: np.ndarray,
) -> Mapping[str, np.ndarray]:
    """Package a set of frames and tracks for use in TAPNet evaluations.

    Given a set of frames and tracks with no query points, use the first
    visible point in each track as the query.

    Args:
      target_occluded: Boolean occlusion flag, of shape [n_tracks, n_frames],
        where True indicates occluded.
      target_points: Position, of shape [n_tracks, n_frames, 2], where each point
        is [x,y] scaled between 0 and 1.
      frames: Video tensor, of shape [n_frames, height, width, 3].  Scaled between
        -1 and 1.

    Returns:
      A dict with the keys:
        video: Video tensor of shape [1, n_frames, height, width, 3]
        query_points: Query points of shape [1, n_queries, 3] where
          each point is [t, y, x] scaled to the range [-1, 1]
        target_points: Target points of shape [1, n_queries, n_frames, 2] where
          each point is [x, y] scaled to the range [-1, 1]
    """

    valid = np.sum(~target_occluded, axis=1) > 0
    target_points = target_points[valid, :]
    target_occluded = target_occluded[valid, :]

    query_points = []
    for i in range(target_points.shape[0]):
        index = np.where(target_occluded[i] == 0)[0][0]
        x, y = target_points[i, index, 0], target_points[i, index, 1]
        query_points.append(np.array([index, y, x]))  # [t, y, x]
    query_points = np.stack(query_points, axis=0)

    return {
        "video": frames[np.newaxis, ...],
        "query_points": query_points[np.newaxis, ...],
        "target_points": target_points[np.newaxis, ...],
        "occluded": target_occluded[np.newaxis, ...],
    }


def create_jhmdb_dataset(
    jhmdb_path: str, resolution: Optional[Tuple[int, int]] = (256, 256)
) -> Iterable[DatasetElement]:
    """JHMDB dataset, including fields required for PCK evaluation."""
    videos = []
    for file in tf.io.gfile.listdir(path.join(gt_dir, "splits")):
        # JHMDB file containing the first split, which is standard for this type of
        # evaluation.
        if not file.endswith("split1.txt"):
            continue

        video_folder = "_".join(file.split("_")[:-2])
        for video in tf.io.gfile.GFile(path.join(gt_dir, "splits", file), "r"):
            video, traintest = video.split()
            video, _ = video.split(".")

            traintest = int(traintest)
            video_path = path.join(video_folder, video)

            if traintest == 2:
                videos.append(video_path)

    if not videos:
        raise ValueError("No JHMDB videos found in directory " + str(jhmdb_path))

    # Shuffle so numbers converge faster.
    random.shuffle(videos)

    for video in videos:
        logging.info(video)
        joints = path.join(gt_dir, "joint_positions", video, "joint_positions.mat")

        if not tf.io.gfile.exists(joints):
            logging.info("skip %s", video)
            continue

        gt_pose = sio.loadmat(tf.io.gfile.GFile(joints, "rb"))["pos_img"]
        gt_pose = np.transpose(gt_pose, [1, 2, 0])
        frames = path.join(gt_dir, "Rename_Images", video, "*.png")
        framefil = tf.io.gfile.glob(frames)
        framefil.sort()

        def read_frame(f):
            im = Image.open(tf.io.gfile.GFile(f, "rb"))
            im = im.convert("RGB")
            im_data = np.array(im.getdata(), np.uint8)
            return im_data.reshape([im.size[1], im.size[0], 3])

        frames = [read_frame(x) for x in framefil]
        frames = np.stack(frames)
        height = frames.shape[1]
        width = frames.shape[2]
        invalid_x = np.logical_or(
            gt_pose[:, 0:1, 0] < 0,
            gt_pose[:, 0:1, 0] >= width,
        )
        invalid_y = np.logical_or(
            gt_pose[:, 0:1, 1] < 0,
            gt_pose[:, 0:1, 1] >= height,
        )
        invalid = np.logical_or(invalid_x, invalid_y)
        invalid = np.tile(invalid, [1, gt_pose.shape[1]])
        invalid = invalid[:, :, np.newaxis].astype(np.float32)
        gt_pose_orig = gt_pose

        if resolution is not None and resolution != frames.shape[1:3]:
            frames = resize_video(frames, resolution)
        frames = frames / (255.0 / 2.0) - 1.0
        queries = gt_pose[:, 0]
        queries = np.concatenate(
            [queries[..., 0:1] * 0, queries[..., ::-1]],
            axis=-1,
        )
        gt_pose = transforms.convert_grid_coordinates(
            gt_pose,
            np.array([width, height]),
            np.array([frames.shape[2], frames.shape[1]]),
        )
        # Set invalid poses to -1 (outside the frame)
        gt_pose = (1.0 - invalid) * gt_pose + invalid * (-1.0)

        if gt_pose.shape[1] < frames.shape[0]:
            # Some videos have pose sequences that are shorter than the frame
            # sequence (usually because the person disappears).  In this case,
            # truncate the video.
            logging.warning("short video!!")
            frames = frames[: gt_pose.shape[1]]

        converted = {
            "video": frames[np.newaxis, ...],
            "query_points": queries[np.newaxis, ...],
            "target_points": gt_pose[np.newaxis, ...],
            "gt_pose": gt_pose[np.newaxis, ...],
            "gt_pose_orig": gt_pose_orig[np.newaxis, ...],
            "occluded": gt_pose[np.newaxis, ..., 0] * 0,
            "fname": video,
            "im_size": np.array([height, width]),
        }
        yield {"jhmdb": converted}


def create_kubric_eval_train_dataset(
    mode: str,
    train_size: Tuple[int, int] = (256, 256),
    max_dataset_size: int = 100,
) -> Iterable[DatasetElement]:
    """Dataset for evaluating performance on Kubric training data."""
    res = dataset.create_point_tracking_dataset(
        split="train",
        train_size=train_size,
        batch_dims=[1],
        shuffle_buffer_size=None,
        repeat=False,
        vflip="vflip" in mode,
        random_crop=False,
    )
    np_ds = tfds.as_numpy(res)

    num_returned = 0
    for data in np_ds:
        if num_returned >= max_dataset_size:
            break
        num_returned += 1
        yield {"kubric": data}


def create_kubric_eval_dataset(
    mode: str, train_size: Tuple[int, int] = (256, 256)
) -> Iterable[DatasetElement]:
    """Dataset for evaluating performance on Kubric val data."""
    res = dataset.create_point_tracking_dataset(
        split="validation",
        train_size=train_size,
        batch_dims=[1],
        shuffle_buffer_size=None,
        repeat=False,
        vflip="vflip" in mode,
        random_crop=False,
    )
    np_ds = tfds.as_numpy(res)

    for data in np_ds:
        yield {"kubric": data}


def create_davis_dataset(
    davis_points_path: str,
    query_mode: str = "strided",
    full_resolution=False,
    resolution: Optional[Tuple[int, int]] = (256, 256),
) -> Iterable[DatasetElement]:
    """Dataset for evaluating performance on DAVIS data."""
    pickle_path = davis_points_path

    with tf.io.gfile.GFile(pickle_path, "rb") as f:
        davis_points_dataset = pickle.load(f)

    if full_resolution:
        ds, _ = tfds.load("davis/full_resolution", split="validation", with_info=True)
        to_iterate = tfds.as_numpy(ds)
    else:
        to_iterate = davis_points_dataset.keys()

    for tmp in to_iterate:
        if full_resolution:
            frames = tmp["video"]["frames"]
            video_name = tmp["metadata"]["video_name"].decode()
        else:
            video_name = tmp
            frames = davis_points_dataset[video_name]["video"]
            if resolution is not None and resolution != frames.shape[1:3]:
                frames = resize_video(frames, resolution)

        frames = frames.astype(np.float32) / 255.0 * 2.0 - 1.0
        target_points = davis_points_dataset[video_name]["points"]
        target_occ = davis_points_dataset[video_name]["occluded"]
        target_points = target_points * np.array([frames.shape[2], frames.shape[1]])

        if query_mode == "strided":
            converted = sample_queries_strided(target_occ, target_points, frames)
        elif query_mode == "first":
            converted = sample_queries_first(target_occ, target_points, frames)
        else:
            raise ValueError(f"Unknown query mode {query_mode}.")

        yield {"davis": converted}


def create_rgb_stacking_dataset(
    robotics_points_path: str,
    query_mode: str = "strided",
    resolution: Optional[Tuple[int, int]] = (256, 256),
) -> Iterable[DatasetElement]:
    """Dataset for evaluating performance on robotics data."""
    pickle_path = robotics_points_path

    with tf.io.gfile.GFile(pickle_path, "rb") as f:
        robotics_points_dataset = pickle.load(f)

    for example in robotics_points_dataset:
        frames = example["video"]
        if resolution is not None and resolution != frames.shape[1:3]:
            frames = resize_video(frames, resolution)
        frames = frames.astype(np.float32) / 255.0 * 2.0 - 1.0
        target_points = example["points"]
        target_occ = example["occluded"]
        target_points = target_points * np.array([frames.shape[2], frames.shape[1]])

        if query_mode == "strided":
            converted = sample_queries_strided(target_occ, target_points, frames)
        elif query_mode == "first":
            converted = sample_queries_first(target_occ, target_points, frames)
        else:
            raise ValueError(f"Unknown query mode {query_mode}.")

        yield {"robotics": converted}


def create_kinetics_dataset(
    kinetics_path: str,
    query_mode: str = "strided",
    resolution: Optional[Tuple[int, int]] = (256, 256),
) -> Iterable[DatasetElement]:
    """Dataset for evaluating performance on Kinetics point tracking."""

    all_paths = tf.io.gfile.glob(path.join(kinetics_path, "*_of_0010.pkl"))
    for pickle_path in all_paths:
        with open(pickle_path, "rb") as f:
            data = pickle.load(f)
            if isinstance(data, dict):
                data = list(data.values())

        # idx = random.randint(0, len(data) - 1)
        for idx in range(len(data)):
            example = data[idx]

            frames = example["video"]

            if isinstance(frames[0], bytes):
                # TAP-Vid is stored and JPEG bytes rather than `np.ndarray`s.
                def decode(frame):
                    byteio = io.BytesIO(frame)
                    img = Image.open(byteio)
                    return np.array(img)

                frames = np.array([decode(frame) for frame in frames])

            if resolution is not None and resolution != frames.shape[1:3]:
                frames = resize_video(frames, resolution)

            frames = frames.astype(np.float32) / 255.0 * 2.0 - 1.0
            target_points = example["points"]
            target_occ = example["occluded"]
            target_points *= np.array([frames.shape[2], frames.shape[1]])

            if query_mode == "strided":
                converted = sample_queries_strided(target_occ, target_points, frames)
            elif query_mode == "first":
                converted = sample_queries_first(target_occ, target_points, frames)
            else:
                raise ValueError(f"Unknown query mode {query_mode}.")

            yield {"kinetics": converted}


def create_csv_dataset(
    dataset_name: str,
    csv_path: str,
    video_base_path: str,
    query_mode: str = "strided",
    resolution: Optional[Tuple[int, int]] = (256, 256),
    max_video_frames: Optional[int] = 1000,
) -> Iterable[DatasetElement]:
    """Create an evaluation iterator out of human annotations and videos.

    Args:
      dataset_name: Name to the dataset.
      csv_path: Path to annotations csv.
      video_base_path: Path to annotated videos.
      query_mode: sample query points from first frame or strided.
      resolution: The video resolution in (height, width).
      max_video_frames: Max length of annotated video.

    Yields:
      Samples for evaluation.
    """
    point_tracks_all = dict()
    with tf.io.gfile.GFile(csv_path, "r") as f:
        reader = csv.reader(f, delimiter=",")
        for row in reader:
            video_id = row[0]
            point_tracks = np.array(row[1:]).reshape(-1, 3)
            if video_id in point_tracks_all:
                point_tracks_all[video_id].append(point_tracks)
            else:
                point_tracks_all[video_id] = [point_tracks]

    for video_id in point_tracks_all:
        if video_id.endswith(".mp4"):
            video_path = path.join(video_base_path, video_id)
        else:
            video_path = path.join(video_base_path, video_id + ".mp4")
        frames = media.read_video(video_path)
        if resolution is not None and resolution != frames.shape[1:3]:
            frames = media.resize_video(frames, resolution)
        frames = frames.astype(np.float32) / 255.0 * 2.0 - 1.0

        point_tracks = np.stack(point_tracks_all[video_id], axis=0)
        point_tracks = point_tracks.astype(np.float32)
        if frames.shape[0] < point_tracks.shape[1]:
            logging.info("Warning: short video!")
            point_tracks = point_tracks[:, : frames.shape[0]]
        point_tracks, occluded = point_tracks[..., 0:2], point_tracks[..., 2]
        occluded = occluded > 0
        target_points = point_tracks * np.array([frames.shape[2], frames.shape[1]])

        num_splits = int(np.ceil(frames.shape[0] / max_video_frames))
        if num_splits > 1:
            print(f"Going to split the video {video_id} into {num_splits}")
        for i in range(num_splits):
            start_index = i * frames.shape[0] // num_splits
            end_index = (i + 1) * frames.shape[0] // num_splits
            sub_occluded = occluded[:, start_index:end_index]
            sub_target_points = target_points[:, start_index:end_index]
            sub_frames = frames[start_index:end_index]

            if query_mode == "strided":
                converted = sample_queries_strided(
                    sub_occluded, sub_target_points, sub_frames
                )
            elif query_mode == "first":
                converted = sample_queries_first(
                    sub_occluded, sub_target_points, sub_frames
                )
            else:
                raise ValueError(f"Unknown query mode {query_mode}.")

            yield {dataset_name: converted}



import cv2


class Colors:
    """Ultralytics color palette https://ultralytics.com/."""

    def __init__(self):
        # fmt: off
        self.palette = [(255, 128, 0), (255, 153, 51), (255, 178, 102), (230, 230, 0), (255, 153, 255),
                            (153, 204, 255), (255, 102, 255), (255, 51, 255), (102, 178, 255), (51, 153, 255),
                            (255, 153, 153), (255, 102, 102), (255, 51, 51), (153, 255, 153), (102, 255, 102),
                            (51, 255, 51), (0, 255, 0), (0, 0, 255), (255, 0, 0), (255, 255, 255)]
        self.palette_index = {i: c for i, c in enumerate(self.palette)}
        # fmt: on

    def __call__(self, indexes):
        """Converts hex color codes to rgb values."""
        return [self.palette_index[idx] for idx in indexes]


# fmt: off
colors = Colors()
kpt_color = colors([16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 9, 9, 9, 9, 9, 9])
skeleton_color = colors([9, 9, 9, 9, 7, 7, 7, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16])

skeleton = [[16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7], [6, 8], [7, 9],
            [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5], [4, 6], [5, 7]]
# fmt: on


class Annotator:
    """A 17 keypoints-based annotator used to define standard styles"""

    def __init__(
        self,
        thickness: int = 1,
        kpts_thickness: int = 4,
        font=cv2.FONT_HERSHEY_DUPLEX,
        font_scale: int = 0.7,
        font_color=(255, 255, 255),
    ) -> None:
        self.thickness = thickness
        self.kpts_thickness = kpts_thickness
        self.font = font
        self.font_scale = font_scale
        self.font_color = font_color

    def draw_bboxes(self, frame, bboxes, labels=[], colors=(0, 0, 255), thickness=None):
        """Draw the bounding boxes (cv2.rectangle)
        bboxes -> [[[x1, y1, x2, y2]], ...]
        thickness -> overrides the self.thickness (int)
        """
        for i, [x1, y1, x2, y2] in enumerate(bboxes):  # index the inner element
            x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
            if len(labels) > i and labels[i]:
                self.draw_text(
                    frame, (x1 + 10, int(y1 - self.font_scale * 20)), str(labels[i])
                )
            cv2.rectangle(
                frame,
                (x1, y1),
                (x2, y2),
                colors,
                thickness if thickness else self.thickness,
            )

    def draw_text(self, frame, coord, text=""):
        cv2.putText(frame, text, coord, self.font, self.font_scale, self.font_color)

    def draw_kpts(self, frame, kpts, colors=kpt_color, thickness=None):
        """Plot the 17 keypoints (cv2.circle)
        kpts -> [[[x, y] * 17], ...] -> shape: (number of human, 17, 2)
        colors -> [(0, 255, 0), ...] -> shape: (17, )
        thickness -> overrides the self.thickness (int)
        """
        for kpt in kpts:
            for i, k in enumerate(kpt):
                cv2.circle(
                    frame,
                    (int(k[0]), int(k[1])),
                    thickness if thickness else self.kpts_thickness,
                    colors[i],
                    -1,
                )

    def draw_skeletons(self, frame, kpts, colors=skeleton_color, thickness=None):
        """Plot the skeletons between two keypoints (cv2.line)
        kpts -> [[[x, y] * 17], ...] -> shape: (number of human, 17, 2)
        colors -> [(0, 255, 0), ...] -> shape: (17, )
        thickness -> overrides the self.thickness (int)
        """

        for kpt in kpts:
            for i, [idx1, idx2] in enumerate(skeleton):
                x1, y1 = kpt[idx1 - 1]
                x2, y2 = kpt[idx2 - 1]
                cv2.line(
                    frame,
                    (int(x1), int(y1)),
                    (int(x2), int(y2)),
                    colors[i],
                    thickness=thickness if thickness else self.thickness,
                )



import contextlib
import inspect
import logging.config
import os
import platform
import re
import subprocess
import sys
import threading
import urllib
import uuid
from datetime import datetime
from pathlib import Path
from types import SimpleNamespace
from typing import Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
import yaml

from juxtapose import __version__

# PyTorch Multi-GPU DDP Constants
RANK = int(os.getenv("RANK", -1))
LOCAL_RANK = int(
    os.getenv("LOCAL_RANK", -1)
)  # https://pytorch.org/docs/stable/elastic/run.html
WORLD_SIZE = int(os.getenv("WORLD_SIZE", 1))

# Other Constants
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # YOLO
# DEFAULT_CFG_PATH = ROOT / "cfg/default.yaml"
NUM_THREADS = min(
    8, max(1, os.cpu_count() - 1)
)  # number of YOLOv5 multiprocessing threads
AUTOINSTALL = (
    str(os.getenv("JUXTAPOSE_AUTOINSTALL", True)).lower() == "true"
)  # global auto-install mode
VERBOSE = (
    str(os.getenv("JUXTAPOSE_VERBOSE", True)).lower() == "true"
)  # global verbose mode
TQDM_BAR_FORMAT = "{l_bar}{bar:10}{r_bar}"  # tqdm bar format
LOGGING_NAME = "JUXTAPOSE"
MACOS, LINUX, WINDOWS = (
    platform.system() == x for x in ["Darwin", "Linux", "Windows"]
)  # environment booleans
ARM64 = platform.machine() in ("arm64", "aarch64")  # ARM64 booleans


# Settings
torch.set_printoptions(linewidth=320, precision=4, profile="default")
np.set_printoptions(
    linewidth=320, formatter={"float_kind": "{:11.5g}".format}
)  # format short g, %precision=5
cv2.setNumThreads(
    0
)  # prevent OpenCV from multithreading (incompatible with PyTorch DataLoader)
os.environ["NUMEXPR_MAX_THREADS"] = str(NUM_THREADS)  # NumExpr max threads
os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"  # for deterministic training
os.environ[
    "TF_CPP_MIN_LOG_LEVEL"
] = "2"  # suppress verbose TF compiler warnings in Colab


class SimpleClass:
    """
    POSE SimpleClass is a base class providing helpful string representation, error reporting, and attribute
    access methods for easier debugging and usage.
    """

    def __str__(self):
        """Return a human-readable string representation of the object."""
        attr = []
        for a in dir(self):
            v = getattr(self, a)
            if not callable(v) and not a.startswith("_"):
                if isinstance(v, SimpleClass):
                    # Display only the module and class name for subclasses
                    s = f"{a}: {v.__module__}.{v.__class__.__name__} object"
                else:
                    s = f"{a}: {repr(v)}"
                attr.append(s)
        return (
            f"{self.__module__}.{self.__class__.__name__} object with attributes:\n\n"
            + "\n".join(attr)
        )

    def __repr__(self):
        """Return a machine-readable string representation of the object."""
        return self.__str__()

    def __getattr__(self, attr):
        """Custom attribute access error message with helpful information."""
        name = self.__class__.__name__
        raise AttributeError(
            f"'{name}' object has no attribute '{attr}'. See valid attributes below.\n{self.__doc__}"
        )


class IterableSimpleNamespace(SimpleNamespace):
    """
    POSE IterableSimpleNamespace is an extension class of SimpleNamespace that adds iterable functionality and
    enables usage with dict() and for loops.
    """

    def __iter__(self):
        """Return an iterator of key-value pairs from the namespace's attributes."""
        return iter(vars(self).items())

    def __str__(self):
        """Return a human-readable string representation of the object."""
        return "\n".join(f"{k}={v}" for k, v in vars(self).items())

    def __getattr__(self, attr):
        """Custom attribute access error message with helpful information."""
        name = self.__class__.__name__
        raise AttributeError(
            f"""
            '{name}' object has no attribute '{attr}'. This may be caused by a modified or out of date JUXTAPOSE
            'default.yaml' file.\nPlease update your code with 'pip install -U juxtapose' and if necessary replace
            with the latest version from
            """
        )

    def get(self, key, default=None):
        """Return the value of the specified key if it exists; otherwise, return the default value."""
        return getattr(self, key, default)


def plt_settings(rcparams=None, backend="Agg"):
    """
    Decorator to temporarily set rc parameters and the backend for a plotting function.

    Usage:
        decorator: @plt_settings({"font.size": 12})
        context manager: with plt_settings({"font.size": 12}):

    Args:
        rcparams (dict): Dictionary of rc parameters to set.
        backend (str, optional): Name of the backend to use. Defaults to 'Agg'.

    Returns:
        (Callable): Decorated function with temporarily set rc parameters and backend. This decorator can be
            applied to any function that needs to have specific matplotlib rc parameters and backend for its execution.
    """

    if rcparams is None:
        rcparams = {"font.size": 11}

    def decorator(func):
        """Decorator to apply temporary rc parameters and backend to a function."""

        def wrapper(*args, **kwargs):
            """Sets rc parameters and backend, calls the original function, and restores the settings."""
            original_backend = plt.get_backend()
            plt.switch_backend(backend)

            with plt.rc_context(rcparams):
                result = func(*args, **kwargs)

            plt.switch_backend(original_backend)
            return result

        return wrapper

    return decorator


def set_logging(name=LOGGING_NAME, verbose=True):
    """Sets up logging for the given name."""
    rank = int(os.getenv("RANK", -1))  # rank in world for Multi-GPU trainings
    level = logging.INFO if verbose and rank in {-1, 0} else logging.ERROR
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {name: {"format": "%(message)s"}},
            "handlers": {
                name: {
                    "class": "logging.StreamHandler",
                    "formatter": name,
                    "level": level,
                }
            },
            "loggers": {name: {"level": level, "handlers": [name], "propagate": False}},
        }
    )


def emojis(string=""):
    """Return platform-dependent emoji-safe version of string."""
    return string.encode().decode("ascii", "ignore") if WINDOWS else string


class EmojiFilter(logging.Filter):
    """
    A custom logging filter class for removing emojis in log messages.

    This filter is particularly useful for ensuring compatibility with Windows terminals
    that may not support the display of emojis in log messages.
    """

    def filter(self, record):
        """Filter logs by emoji unicode characters on windows."""
        record.msg = emojis(record.msg)
        return super().filter(record)


# Set logger
set_logging(LOGGING_NAME, verbose=VERBOSE)  # run before defining LOGGER
LOGGER = logging.getLogger(
    LOGGING_NAME
)  # define globally (used in train.py, val.py, detect.py, etc.)
if WINDOWS:  # emoji-safe logging
    LOGGER.addFilter(EmojiFilter())


class ThreadingLocked:
    """
    A decorator class for ensuring thread-safe execution of a function or method.
    This class can be used as a decorator to make sure that if the decorated function
    is called from multiple threads, only one thread at a time will be able to execute the function.

    Attributes:
        lock (threading.Lock): A lock object used to manage access to the decorated function.

    Usage:
        @ThreadingLocked()
        def my_function():
            # Your code here
            pass
    """

    def __init__(self):
        self.lock = threading.Lock()

    def __call__(self, f):
        from functools import wraps

        @wraps(f)
        def decorated(*args, **kwargs):
            with self.lock:
                return f(*args, **kwargs)

        return decorated


def yaml_save(file="data.yaml", data=None):
    """
    Save YAML data to a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        data (dict): Data to save in YAML format.

    Returns:
        (None): Data is saved to the specified file.
    """
    if data is None:
        data = {}
    file = Path(file)
    if not file.parent.exists():
        # Create parent directories if they don't exist
        file.parent.mkdir(parents=True, exist_ok=True)

    # Convert Path objects to strings
    for k, v in data.items():
        if isinstance(v, Path):
            data[k] = str(v)

    # Dump data to file in YAML format
    with open(file, "w") as f:
        yaml.safe_dump(data, f, sort_keys=False, allow_unicode=True)


def yaml_load(file="data.yaml", append_filename=False):
    """
    Load YAML data from a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        append_filename (bool): Add the YAML filename to the YAML dictionary. Default is False.

    Returns:
        (dict): YAML data and file name.
    """
    with open(file, errors="ignore", encoding="utf-8") as f:
        s = f.read()  # string

        # Remove special characters
        if not s.isprintable():
            s = re.sub(
                r"[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\uD7FF\uE000-\uFFFD\U00010000-\U0010ffff]+",
                "",
                s,
            )

        # Add YAML filename to dict and return
        return (
            {**yaml.safe_load(s), "yaml_file": str(file)}
            if append_filename
            else yaml.safe_load(s)
        )


def yaml_print(yaml_file: Union[str, Path, dict]) -> None:
    """
    Pretty prints a yaml file or a yaml-formatted dictionary.

    Args:
        yaml_file: The file path of the yaml file or a yaml-formatted dictionary.

    Returns:
        None
    """
    yaml_dict = (
        yaml_load(yaml_file) if isinstance(yaml_file, (str, Path)) else yaml_file
    )
    dump = yaml.dump(yaml_dict, sort_keys=False, allow_unicode=True)
    LOGGER.info(f"Printing '{colorstr('bold', 'black', yaml_file)}'\n\n{dump}")


# Default configuration
# DEFAULT_CFG_DICT = yaml_load(DEFAULT_CFG_PATH)
# for k, v in DEFAULT_CFG_DICT.items():
#     if isinstance(v, str) and v.lower() == "none":
#         DEFAULT_CFG_DICT[k] = None
# DEFAULT_CFG_KEYS = DEFAULT_CFG_DICT.keys()
# DEFAULT_CFG = IterableSimpleNamespace(**DEFAULT_CFG_DICT)


def is_colab():
    """
    Check if the current script is running inside a Google Colab notebook.

    Returns:
        (bool): True if running inside a Colab notebook, False otherwise.
    """
    return "COLAB_RELEASE_TAG" in os.environ or "COLAB_BACKEND_VERSION" in os.environ


def is_kaggle():
    """
    Check if the current script is running inside a Kaggle kernel.

    Returns:
        (bool): True if running inside a Kaggle kernel, False otherwise.
    """
    return (
        os.environ.get("PWD") == "/kaggle/working"
        and os.environ.get("KAGGLE_URL_BASE") == "https://www.kaggle.com"
    )


def is_jupyter():
    """
    Check if the current script is running inside a Jupyter Notebook.
    Verified on Colab, Jupyterlab, Kaggle, Paperspace.

    Returns:
        (bool): True if running inside a Jupyter Notebook, False otherwise.
    """
    with contextlib.suppress(Exception):
        from IPython import get_ipython

        return get_ipython() is not None
    return False


def is_docker() -> bool:
    """
    Determine if the script is running inside a Docker container.

    Returns:
        (bool): True if the script is running inside a Docker container, False otherwise.
    """
    file = Path("/proc/self/cgroup")
    if file.exists():
        with open(file) as f:
            return "docker" in f.read()
    else:
        return False


def is_online() -> bool:
    """
    Check internet connectivity by attempting to connect to a known online host.

    Returns:
        (bool): True if connection is successful, False otherwise.
    """
    import socket

    for host in "1.1.1.1", "8.8.8.8", "223.5.5.5":  # Cloudflare, Google, AliDNS:
        try:
            test_connection = socket.create_connection(address=(host, 53), timeout=2)
        except (socket.timeout, socket.gaierror, OSError):
            continue
        else:
            # If the connection was successful, close it to avoid a ResourceWarning
            test_connection.close()
            return True
    return False


ONLINE = is_online()


def is_pip_package(filepath: str = __name__) -> bool:
    """
    Determines if the file at the given filepath is part of a pip package.

    Args:
        filepath (str): The filepath to check.

    Returns:
        (bool): True if the file is part of a pip package, False otherwise.
    """
    import importlib.util

    # Get the spec for the module
    spec = importlib.util.find_spec(filepath)

    # Return whether the spec is not None and the origin is not None (indicating it is a package)
    return spec is not None and spec.origin is not None


def is_dir_writeable(dir_path: Union[str, Path]) -> bool:
    """
    Check if a directory is writeable.

    Args:
        dir_path (str | Path): The path to the directory.

    Returns:
        (bool): True if the directory is writeable, False otherwise.
    """
    return os.access(str(dir_path), os.W_OK)


def is_pytest_running():
    """
    Determines whether pytest is currently running or not.

    Returns:
        (bool): True if pytest is running, False otherwise.
    """
    return (
        ("PYTEST_CURRENT_TEST" in os.environ)
        or ("pytest" in sys.modules)
        or ("pytest" in Path(sys.argv[0]).stem)
    )


def is_github_actions_ci() -> bool:
    """
    Determine if the current environment is a GitHub Actions CI Python runner.

    Returns:
        (bool): True if the current environment is a GitHub Actions CI Python runner, False otherwise.
    """
    return (
        "GITHUB_ACTIONS" in os.environ
        and "RUNNER_OS" in os.environ
        and "RUNNER_TOOL_CACHE" in os.environ
    )


def is_git_dir():
    """
    Determines whether the current file is part of a git repository.
    If the current file is not part of a git repository, returns None.

    Returns:
        (bool): True if current file is part of a git repository.
    """
    return get_git_dir() is not None


def get_git_dir():
    """
    Determines whether the current file is part of a git repository and if so, returns the repository root directory.
    If the current file is not part of a git repository, returns None.

    Returns:
        (Path | None): Git root directory if found or None if not found.
    """
    for d in Path(__file__).parents:
        if (d / ".git").is_dir():
            return d
    return None  # no .git dir found


def get_git_origin_url():
    """
    Retrieves the origin URL of a git repository.

    Returns:
        (str | None): The origin URL of the git repository.
    """
    if is_git_dir():
        with contextlib.suppress(subprocess.CalledProcessError):
            origin = subprocess.check_output(
                ["git", "config", "--get", "remote.origin.url"]
            )
            return origin.decode().strip()
    return None  # if not git dir or on error


def get_git_branch():
    """
    Returns the current git branch name. If not in a git repository, returns None.

    Returns:
        (str | None): The current git branch name.
    """
    if is_git_dir():
        with contextlib.suppress(subprocess.CalledProcessError):
            origin = subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"]
            )
            return origin.decode().strip()
    return None  # if not git dir or on error


def get_default_args(func):
    """Returns a dictionary of default arguments for a function.

    Args:
        func (callable): The function to inspect.

    Returns:
        (dict): A dictionary where each key is a parameter name, and each value is the default value of that parameter.
    """
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def get_user_config_dir(sub_dir="juxtapose"):
    """
    Get the user config directory.

    Args:
        sub_dir (str): The name of the subdirectory to create.

    Returns:
        (Path): The path to the user config directory.
    """
    # Return the appropriate config directory for each operating system
    if WINDOWS:
        path = Path.home() / "AppData" / "Roaming" / sub_dir
    elif MACOS:  # macOS
        path = Path.home() / "Library" / "Application Support" / sub_dir
    elif LINUX:
        path = Path.home() / ".config" / sub_dir
    else:
        raise ValueError(f"Unsupported operating system: {platform.system()}")

    # GCP and AWS lambda fix, only /tmp is writeable
    if not is_dir_writeable(str(path.parent)):
        path = Path("/tmp") / sub_dir
        LOGGER.warning(
            f"WARNING ⚠️ user config directory is not writeable, defaulting to '{path}'."
        )

    # Create the subdirectory if it does not exist
    path.mkdir(parents=True, exist_ok=True)

    return path


USER_CONFIG_DIR = Path(
    os.getenv("YOLO_CONFIG_DIR", get_user_config_dir())
)  # POSE settings dir
SETTINGS_YAML = USER_CONFIG_DIR / "settings.yaml"


def colorstr(*input):
    """Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')."""
    *args, string = (
        input if len(input) > 1 else ("blue", "bold", input[0])
    )  # color arguments, string
    colors = {
        "black": "\033[30m",  # basic colors
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "white": "\033[37m",
        "bright_black": "\033[90m",  # bright colors
        "bright_red": "\033[91m",
        "bright_green": "\033[92m",
        "bright_yellow": "\033[93m",
        "bright_blue": "\033[94m",
        "bright_magenta": "\033[95m",
        "bright_cyan": "\033[96m",
        "bright_white": "\033[97m",
        "end": "\033[0m",  # misc
        "bold": "\033[1m",
        "underline": "\033[4m",
    }
    return "".join(colors[x] for x in args) + f"{string}" + colors["end"]


class TryExcept(contextlib.ContextDecorator):
    """YOLOv8 TryExcept class. Usage: @TryExcept() decorator or 'with TryExcept():' context manager."""

    def __init__(self, msg="", verbose=True):
        """Initialize TryExcept class with optional message and verbosity settings."""
        self.msg = msg
        self.verbose = verbose

    def __enter__(self):
        """Executes when entering TryExcept context, initializes instance."""
        pass

    def __exit__(self, exc_type, value, traceback):
        """Defines behavior when exiting a 'with' block, prints error message if necessary."""
        if self.verbose and value:
            print(emojis(f"{self.msg}{': ' if self.msg else ''}{value}"))
        return True


def threaded(func):
    """Multi-threads a target function and returns thread. Usage: @threaded decorator."""

    def wrapper(*args, **kwargs):
        """Multi-threads a given function and returns the thread."""
        thread = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
        thread.start()
        return thread

    return wrapper


def update_dict_recursive(d, u):
    """
    Recursively updates the dictionary `d` with the key-value pairs from the dictionary `u` without overwriting
    entire sub-dictionaries. Note that function recursion is intended and not a problem, as this allows for updating
    nested dictionaries at any arbitrary depth.

    Args:
        d (dict): The dictionary to be updated.
        u (dict): The dictionary to update `d` with.

    Returns:
        (dict): The recursively updated dictionary.
    """
    for k, v in u.items():
        d[k] = update_dict_recursive(d.get(k, {}), v) if isinstance(v, dict) else v
    return d


def deprecation_warn(arg, new_arg, version=None):
    """Issue a deprecation warning when a deprecated argument is used, suggesting an updated argument."""
    if not version:
        version = float(__version__[:3]) + 0.2  # deprecate after 2nd major release
    LOGGER.warning(
        f"WARNING ⚠️ '{arg}' is deprecated and will be removed in 'juxtapose {version}' in the future. "
        f"Please use '{new_arg}' instead."
    )


def clean_url(url):
    """Strip auth from URL, i.e. https://url.com/file.txt?auth -> https://url.com/file.txt."""
    url = (
        Path(url).as_posix().replace(":/", "://")
    )  # Pathlib turns :// -> :/, as_posix() for Windows
    return urllib.parse.unquote(url).split("?")[
        0
    ]  # '%2F' to '/', split https://url.com/file.txt?auth


def url2file(url):
    """Convert URL to filename, i.e. https://url.com/file.txt?auth -> file.txt."""
    return Path(clean_url(url)).name


def get_time():
    # current date and time
    return datetime.now().strftime("%Y%m%d-%H%M%S")


DEFAULT_COLOR_PALETTE = [
    "#a351fb",
    "#e6194b",
    "#3cb44b",
    "#ffe119",
    "#0082c8",
    "#f58231",
    "#911eb4",
    "#46f0f0",
    "#f032e6",
    "#d2f53c",
    "#fabebe",
    "#008080",
    "#e6beff",
    "#aa6e28",
    "#fffac8",
    "#800000",
    "#aaffc3",
]


# Run below code on utils init ------------------------------------------------------------------------------------

# Check first-install steps
PREFIX = colorstr("JUXTAPOSE: ")
ENVIRONMENT = (
    "Colab"
    if is_colab()
    else "Kaggle"
    if is_kaggle()
    else "Jupyter"
    if is_jupyter()
    else "Docker"
    if is_docker()
    else platform.system()
)
TESTS_RUNNING = is_pytest_running() or is_github_actions_ci()

# Apply monkey patches if the script is being run from within the parent directory of the script's location
from .patches import imread, imshow, imwrite

# torch.save = torch_save
if (
    Path(inspect.stack()[0].filename).parent.parent.as_posix()
    in inspect.stack()[-1].filename
):
    cv2.imread, cv2.imwrite, cv2.imshow = imread, imwrite, imshow



import cv2
from dataclasses import replace
import numpy as np
from typing import Tuple
from .core import Detections

from enum import Enum


class Position(Enum):
    """
    Enum representing the position of an anchor point.
    """

    CENTER = "CENTER"
    CENTER_LEFT = "CENTER_LEFT"
    CENTER_RIGHT = "CENTER_RIGHT"
    TOP_CENTER = "TOP_CENTER"
    TOP_LEFT = "TOP_LEFT"
    TOP_RIGHT = "TOP_RIGHT"
    BOTTOM_LEFT = "BOTTOM_LEFT"
    BOTTOM_CENTER = "BOTTOM_CENTER"
    BOTTOM_RIGHT = "BOTTOM_RIGHT"
    CENTER_OF_MASS = "CENTER_OF_MASS"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def clip_boxes(xyxy: np.ndarray, resolution_wh: Tuple[int, int]) -> np.ndarray:
    """
    Clips bounding boxes coordinates to fit within the frame resolution.

    Args:
        xyxy (np.ndarray): A numpy array of shape `(N, 4)` where each
            row corresponds to a bounding box in
        the format `(x_min, y_min, x_max, y_max)`.
        resolution_wh (Tuple[int, int]): A tuple of the form `(width, height)`
            representing the resolution of the frame.

    Returns:
        np.ndarray: A numpy array of shape `(N, 4)` where each row
            corresponds to a bounding box with coordinates clipped to fit
            within the frame resolution.
    """
    result = np.copy(xyxy)
    width, height = resolution_wh
    result[:, [0, 2]] = result[:, [0, 2]].clip(0, width)
    result[:, [1, 3]] = result[:, [1, 3]].clip(0, height)
    return result


def polygon_to_mask(polygon: np.ndarray, resolution_wh: Tuple[int, int]) -> np.ndarray:
    """Generate a mask from a polygon.

    Args:
        polygon (np.ndarray): The polygon for which the mask should be generated,
            given as a list of vertices.
        resolution_wh (Tuple[int, int]): The width and height of the desired resolution.

    Returns:
        np.ndarray: The generated 2D mask, where the polygon is marked with
            `1`'s and the rest is filled with `0`'s.
    """
    width, height = resolution_wh
    mask = np.zeros((height, width))

    cv2.fillPoly(mask, [polygon], color=1)
    return mask


class PolygonZone:
    """
    A class for defining a polygon-shaped zone within a frame for detecting objects.

    Attributes:
        polygon (np.ndarray): A polygon represented by a numpy array of shape
            `(N, 2)`, containing the `x`, `y` coordinates of the points.
        frame_resolution_wh (Tuple[int, int]): The frame resolution (width, height)
        triggering_position (Position): The position within the bounding
            box that triggers the zone (default: Position.BOTTOM_CENTER)
        current_count (int): The current count of detected objects within the zone
        mask (np.ndarray): The 2D bool mask for the polygon zone
    """

    def __init__(
        self,
        polygon: np.ndarray,
        frame_resolution_wh: Tuple[int, int],
        triggering_position: Position = Position.BOTTOM_CENTER,
    ):
        self.polygon = polygon.astype(int)
        self.frame_resolution_wh = frame_resolution_wh
        self.triggering_position = triggering_position
        self.current_count = 0

        width, height = frame_resolution_wh
        self.mask = polygon_to_mask(
            polygon=polygon, resolution_wh=(width + 1, height + 1)
        )

    def trigger(self, detections: Detections) -> np.ndarray:
        """
        Determines if the detections are within the polygon zone.

        Parameters:
            detections (Detections): The detections
                to be checked against the polygon zone

        Returns:
            np.ndarray: A boolean numpy array indicating
                if each detection is within the polygon zone
        """

        clipped_xyxy = clip_boxes(
            xyxy=detections.xyxy, resolution_wh=self.frame_resolution_wh
        )
        clipped_detections = replace(detections, xyxy=clipped_xyxy)
        clipped_anchors = np.ceil(
            clipped_detections.get_anchor_coordinates(anchor=self.triggering_position)
        ).astype(int)
        is_in_zone = self.mask[clipped_anchors[:, 1], clipped_anchors[:, 0]]
        self.current_count = int(np.sum(is_in_zone))
        return is_in_zone.astype(bool)



from .groundingdino import GroundingDino
from .rtmdet import RTMDet
from .yolov8 import YOLOv8

DET_MAP = {"rtmdet": RTMDet, "groundingdino": GroundingDino, "yolov8": YOLOv8}


def get_detector(model: str = "rtmdet-l", *args, **kwargs):
    if model.startswith("groundingdino"):
        return DET_MAP["groundingdino"]("", *args, **kwargs)
    elif model.startswith("yolov8"):
        del kwargs["captions"]
        return DET_MAP["yolov8"](model[-1], *args, **kwargs)
    else:
        del kwargs["captions"]
        type, size = model.split("-")
        return DET_MAP[type](size, *args, **kwargs)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities for transforming image coordinates."""

from typing import Sequence

import chex
import numpy as np


def convert_grid_coordinates(
    coords: chex.Array,
    input_grid_size: Sequence[int],
    output_grid_size: Sequence[int],
    coordinate_format: str = 'xy',
) -> chex.Array:
  """Convert image coordinates between image grids of different sizes.

  By default, it assumes that the image corners are aligned.  Therefore,
  it adds .5 (since (0,0) is assumed to be the center of the upper-left grid
  cell), multiplies by the size ratio, and then subtracts .5.

  Args:
    coords: The coordinates to be converted.  It is of shape [..., 2] if
      coordinate_format is 'xy' or [..., 3] if coordinate_format is 'tyx'.
    input_grid_size: The size of the image/grid that the coordinates currently
      are with respect to.  This is a 2-tuple of the format [width, height]
      if coordinate_format is 'xy' or a 3-tuple of the format
      [num_frames, height, width] if coordinate_format is 'tyx'.
    output_grid_size: The size of the target image/grid that you want the
      coordinates to be with respect to.  This is a 2-tuple of the format
      [width, height] if coordinate_format is 'xy' or a 3-tuple of the format
      [num_frames, height, width] if coordinate_format is 'tyx'.
    coordinate_format: Which format the coordinates are in.  This can be one
      of 'xy' (the default) or 'tyx', which are the only formats used in this
      project.

  Returns:
    The transformed coordinates, of the same shape as coordinates.

  Raises:
    ValueError: if coordinates don't match the given format.
  """
  if isinstance(input_grid_size, tuple):
    input_grid_size = np.array(input_grid_size)
  if isinstance(output_grid_size, tuple):
    output_grid_size = np.array(output_grid_size)

  if coordinate_format == 'xy':
    if input_grid_size.shape[0] != 2 or output_grid_size.shape[0] != 2:
      raise ValueError(
          'If coordinate_format is xy, the shapes must be length 2.')
  elif coordinate_format == 'tyx':
    if input_grid_size.shape[0] != 3 or output_grid_size.shape[0] != 3:
      raise ValueError(
          'If coordinate_format is tyx, the shapes must be length 3.')
    if input_grid_size[0] != output_grid_size[0]:
      raise ValueError('converting frame count is not supported.')
  else:
    raise ValueError('Recognized coordinate formats are xy and tyx.')

  position_in_grid = coords
  position_in_grid = position_in_grid * output_grid_size / input_grid_size

  return position_in_grid



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utils functions for TSM."""

from typing import Tuple

import chex
import jax
import jax.numpy as jnp


def prepare_inputs(inputs: chex.Array) -> Tuple[jnp.ndarray, str, int]:
  """Deduces input mode for TSM."""
  # Deduce if we run on TPU based on input shape.
  if len(inputs.shape) == 5:
    # Input is given in the standard [B, T, H, W, 3] format.
    tsm_mode = 'gpu'
    num_frames = inputs.shape[1]
    inputs = jnp.reshape(inputs, [-1] + list(inputs.shape[2:]))
  else:
    # Input is given in the [T * B, H, W, 3] format.
    tsm_mode = 'tpu'
    num_frames = None
  return inputs, tsm_mode, num_frames


def prepare_outputs(
    outputs: chex.Array,
    tsm_mode: str,
    num_frames: int,
    reduce_mean: bool = True,
) -> jnp.ndarray:
  """Processes output of TSM to undo the merging of batch and time."""
  # Get the shape without the batch/time dimension (for TSM batch and time are
  # merged in the first dimension).
  shape_no_bt = list(outputs.shape[1:])
  if tsm_mode == 'tpu':
    # Outputs are of the shape [num_frames * B, ..., n_channels]
    outputs = jnp.reshape(outputs, [num_frames, -1] + shape_no_bt)
    if reduce_mean:
      # We average over time and space.
      outputs = jnp.mean(
          outputs, axis=[0] + list(range(2,
                                         len(shape_no_bt) + 1)))
    else:
      outputs = jnp.transpose(
          outputs, axes=[1, 0] + list(range(2,
                                            len(shape_no_bt) + 2)))
  elif tsm_mode == 'gpu':
    # Outputs are of the shape [B * num_frames, ..., n_channels].
    outputs = jnp.reshape(outputs, [-1, num_frames] + shape_no_bt)
    if reduce_mean:
      outputs = jnp.mean(
          outputs, axis=[1] + list(range(2,
                                         len(shape_no_bt) + 1)))
  elif tsm_mode.startswith('deflated'):
    # In deflated mode, outputs are already in the right format.
    pass
  else:
    raise ValueError('`tsm_mode` should be \'tpu\' or \'gpu\' or '
                     f'\'deflated_0.x\' ({tsm_mode} given)')
  return outputs  # pytype: disable=bad-return-type  # numpy-scalars


def apply_temporal_shift(
    x: chex.Array,
    tsm_mode: str,
    num_frames: int,
    channel_shift_fraction: float = 0.125,
) -> jnp.ndarray:
  """Performs a temporal shift: https://arxiv.org/abs/1811.08383 with mode."""
  if tsm_mode == 'tpu':
    outputs = temporal_shift_tpu(x, num_frames, channel_shift_fraction)
  elif tsm_mode == 'gpu':
    outputs = temporal_shift_gpu(x, num_frames, channel_shift_fraction)
  elif tsm_mode.startswith('deflated'):
    alpha = float(tsm_mode.split('_')[1])
    outputs = temporal_shift_image_mode(x, channel_shift_fraction, alpha)
  else:
    raise ValueError('`tsm_mode` should be \'tpu\' or \'gpu\' or '
                     f'\'deflated_0.x\' ({tsm_mode} given)')
  return outputs


def temporal_shift_image_mode(x, channel_shift_fraction=0.125, alpha=0.3):
  """Temporal shift applied on single image (to emulate a fixed video)."""
  # B, H, W, C = batch_size, im_height, im_width, channels.
  # Input is (B, H, W, C).
  orig_shp = tuple(x.shape)
  n_channels = orig_shp[-1]
  n_shift = int(n_channels * channel_shift_fraction)
  # Alpha emulates the effect of the padding when using a single frame.
  shifted_backward = alpha * x[:, :, :, -n_shift:]
  shifted_forward = alpha * x[:, :, :, :n_shift]
  no_shift = x[:, :, :, n_shift:-n_shift]
  shifted_x = jnp.concatenate([shifted_backward, no_shift, shifted_forward],
                              axis=3)
  return shifted_x


def temporal_shift_gpu(
    x: chex.Array,
    num_frames: int,
    channel_shift_fraction: float = 0.125,
) -> jnp.ndarray:
  """Performs a temporal shift: https://arxiv.org/abs/1811.08383."""
  # B, T, H, W, C = batch_size, num_frames, im_height, im_width, channels.
  # Input is (B * T, H, W, C).
  orig_shp = tuple(x.shape)
  reshaped_x = jnp.reshape(x, (-1, num_frames) + orig_shp[1:])
  n_channels = orig_shp[-1]
  n_shift = int(n_channels * channel_shift_fraction)

  new_shp = tuple(reshaped_x.shape)

  # shifted_backward = reshaped_x[:, 1:, :, :, -n_shift:].
  shifted_backward = jax.lax.slice(
      reshaped_x, (0, 1, 0, 0, new_shp[4] - n_shift),
      (new_shp[0], new_shp[1], new_shp[2], new_shp[3], new_shp[4]))
  shifted_backward_padding = ((0, 0), (0, 1), (0, 0), (0, 0), (0, 0))
  shifted_backward = jnp.pad(shifted_backward, shifted_backward_padding)

  # shifted_forward = reshaped_x[:, :-1, :, :, :n_shift].
  shifted_forward = jax.lax.slice(
      reshaped_x, (0, 0, 0, 0, 0),
      (new_shp[0], new_shp[1] - 1, new_shp[2], new_shp[3], n_shift))
  shifted_forward_padding = ((0, 0), (1, 0), (0, 0), (0, 0), (0, 0))
  shifted_forward = jnp.pad(shifted_forward, shifted_forward_padding)

  no_shift = reshaped_x[:, :, :, :, n_shift:-n_shift]
  shifted_x = jnp.concatenate([shifted_backward, no_shift, shifted_forward],
                              axis=4)
  return jnp.reshape(shifted_x, (-1,) + orig_shp[1:])


def temporal_shift_tpu(
    x: chex.Array,
    num_frames: int,
    channel_shift_fraction: float = 0.125,
) -> jnp.ndarray:
  """Performs a temporal shift: https://arxiv.org/abs/1811.08383.

    TPU optimized version of TSM. Reshape is avoided by having the images
    reshaped in [T * B, :] so that frames corresponding to same time frame in
    videos are contiguous in memory. Finally, to avoid concatenate that prevent
    some fusion from happening we simply sum masked version of the features.
  Args:
    x: Input expected to be [T * B, H, W, C] (where the batch has been reshaped
      from a time major version of the input).
    num_frames: number of frames T per video.
    channel_shift_fraction: fraction of the channel to shift forward and
      backward.

  Returns:
      The temporal shifted version of x.
  """
  # B, T, H, W, C = batch_size, num_frames, im_height, im_width, channels.
  # Input is (T * B, H, W, C).
  original_dtype = x.dtype
  original_shape = list(x.shape)

  batch_size = int(original_shape[0] / num_frames)
  n_channels = int(original_shape[-1])
  n_shift = int(n_channels * channel_shift_fraction)

  # Cast to bfloat16.
  x = x.astype(jnp.bfloat16)

  # For the following, assume that x has 3 channels [x1, x2, x3] and n_shift=1.
  # Shift backward, we first pad by zeros [x1, x2, x3, 0, 0].
  orig_shp = list(x.shape)

  shifted_backward_padding = ((0, batch_size, 0), (0, 0, 0), (0, 0, 0),
                              (0, n_channels - n_shift, 0))
  x_backward_padding = jax.lax.pad(
      x,
      padding_value=jnp.bfloat16(0.),
      padding_config=shifted_backward_padding)
  # The following shift gets to [x3^+1, 0, 0] (where +1 means from the future).
  shifted_backward = jax.lax.slice(x_backward_padding,
                                   (batch_size, 0, 0, n_channels - n_shift),
                                   (orig_shp[0] + batch_size, orig_shp[1],
                                    orig_shp[2], 2 * n_channels - n_shift))
  # Shift forward, we first pad by zeros [0, 0, x1, x2, x3].
  shifted_forward_padding = ((batch_size, 0, 0), (0, 0, 0), (0, 0, 0),
                             (n_channels - n_shift, 0, 0))
  x_forward_padding = jax.lax.pad(
      x, padding_value=jnp.bfloat16(0.), padding_config=shifted_forward_padding)
  # The following shift gets to [0, 0, x1^-1] (where -1 means from the past).
  shifted_forward = jax.lax.slice(
      x_forward_padding, (0, 0, 0, 0),
      (orig_shp[0], orig_shp[1], orig_shp[2], n_channels))
  # No shift is in the middle, this gets [0, x2, 0].
  mask_noshift = (jnp.reshape((jnp.arange(n_channels) >= n_shift) &
                              (jnp.arange(n_channels) < n_channels - n_shift),
                              (1, 1, 1, -1))).astype(jnp.bfloat16)
  no_shift = mask_noshift * x
  # By summing everything together, we end up with [x3^+1, x2, x1^-1].
  # Note: channels have been reordered but that doesn't matter for the model.
  shifted_x = shifted_backward + shifted_forward + no_shift

  return shifted_x.astype(original_dtype)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A Jaxline script for training and evaluating TAPNet."""

import sys
from typing import Callable, Dict, Iterable, Iterator, Mapping, Optional, Tuple

from absl import app
from absl import flags
from absl import logging
import chex
import haiku as hk
import jax
import jax.numpy as jnp

from jaxline import experiment
from jaxline import platform
from jaxline import utils
from kubric.challenges.point_tracking import dataset
from ml_collections import config_dict

import numpy as np
import optax
import tensorflow as tf
import tensorflow_datasets as tfds

from juxtapose.trackers.tapnet import supervised_point_prediction
from juxtapose.trackers.tapnet import tapir_model
from juxtapose.trackers.tapnet import tapnet_model
from juxtapose.trackers.tapnet import task
from juxtapose.trackers.tapnet.utils import experiment_utils as exputils


class Experiment(experiment.AbstractExperiment):
    """TAPNet experiment.

    This constructs a set of tasks which are used to compute gradients, as well as
    a set of datasets which are passed into every task.  After the tasks compute
    gradients using the data, this experiment will aggregate and apply those
    gradients using an optimizer.

    This experiment is organized around the needs of Haiku's multi transform:
    each task specifies its own forward function, but different tasks need to
    share parameters.  Therefore, the design of this experiment is that each
    task has a forward function, but the global Experiment class is responsible
    for creating the multi transform out of these.  The Experiment class keeps
    track of the variables.  Each task doesn't directly call its own forward
    function.  Instead receives a "wrapped" forward function which contains
    the Haiku state, which injected (via closure) by the Experiment class.

    To do its job, multi transform needs two things: a function
    which initializes all variables, and the set of forward functions that need
    to be called separately, one for each task.  The  latter is just the set of
    forward functions for each task.  The initialization function is
    just a function that calls every forward function sequentially.
    """

    # A map from object properties that will be checkpointed to their name
    # in a checkpoint. Currently we assume that these are all sharded
    # device arrays.
    CHECKPOINT_ATTRS = {
        "_params": "params",
        "_state": "state",
        "_opt_state": "opt_state",
    }

    def __init__(
        self,
        mode: str,
        init_rng: jnp.ndarray,
        config: config_dict.ConfigDict,
    ):
        """Initializes the experiment.

        This includes constructing all of the requested tasks, creating the Haiku
        transforms, and pmapping the update function that will be used throughout
        training.  Currently the two supported training tasks are 'selfsup_tracks'
        and 'kubric'.

        Args:
          mode: either 'train' (for training) or one of the recognized eval modes
            (see Experiment.evaluate).
          init_rng: jax.random.PRNGKey for random number generation.
          config: config options.  See configs/tapnet_config.py for an example.
        """

        super().__init__(mode=mode, init_rng=init_rng)

        self.mode = mode
        self.init_rng = init_rng
        self.config = config

        # Checkpointed experiment state.
        self._params = None
        self._state = None
        self._opt_state = None

        self._optimizer = None

        # Input pipelines.
        self._train_input = None
        self._eval_input = None

        self.point_prediction = supervised_point_prediction.SupervisedPointPrediction(
            config, **config.supervised_point_prediction_kwargs
        )

        def forward(*args, is_training=True, **kwargs):
            shared_modules = self._construct_shared_modules()
            return self.point_prediction.forward_fn(
                *args,
                shared_modules=shared_modules,
                is_training=is_training,
                **kwargs,
            )

        self._transform = hk.transform_with_state(forward)

        # NOTE: We "donate" the `params, state, opt_state` arguments which allows
        # JAX (on some backends) to reuse the device memory associated with these
        # inputs to store the outputs of our function (which also start with
        # `params, state, opt_state`).
        self._update_func = jax.pmap(
            self._update_func, axis_name="i", donate_argnums=(0, 1, 2)
        )

    def _construct_shared_modules(self) -> Mapping[str, task.SharedModule]:
        """Constructs the TAPNet module which is used for all tasks.

        More generally, these are Haiku modules that are passed to all tasks so that
        weights are shared across tasks.

        Returns:
          A dict with a single key 'tapnet_model' containing the tapnet model.
        """
        shared_module_constructors = {
            "tapnet_model": tapnet_model.TAPNet,
            "tapir_model": tapir_model.TAPIR,
        }
        shared_modules = {}

        for shared_mod_name in self.config.shared_modules.shared_module_names:
            ctor = shared_module_constructors[shared_mod_name]
            kwargs = self.config.shared_modules[shared_mod_name + "_kwargs"]
            shared_modules[shared_mod_name] = ctor(**kwargs)
        return shared_modules

    #  _             _
    # | |_ _ __ __ _(_)_ __
    # | __| '__/ _` | | '_ \
    # | |_| | | (_| | | | | |
    #  \__|_|  \__,_|_|_| |_|
    #
    def step(  # pytype: disable=signature-mismatch  # numpy-scalars
        self,
        global_step: chex.Array,
        rng: chex.PRNGKey,
        *unused_args,
        **unused_kwargs,
    ) -> Dict[str, chex.Array]:
        """See base class."""

        if self._train_input is None:
            self._initialize_train()

        inputs = next(self._train_input)

        self._params, self._state, self._opt_state, scalars = self._update_func(
            self._params,
            self._state,
            self._opt_state,
            inputs,
            rng,
            global_step,
        )

        scalars = utils.get_first(scalars)

        if ((utils.get_first(global_step) + 1) % self.config.evaluate_every) == 0:
            for mode in self.config.eval_modes:
                eval_scalars = self.evaluate(global_step, rng=rng, mode=mode)
                scalars.update(eval_scalars)

        return scalars

    def _initialize_train(self):
        """Initializes the training parameters using the first training input.

        self._multi_transform.init will call the forward_fn, which allows
        Haiku to identify the params that need to be initialized.
        """
        self._train_input = utils.py_prefetch(self._build_train_input)

        total_steps = self.config.training.n_training_steps
        # Scale by the (negative) learning rate.
        # Check we haven't already restored params
        if self._params is None:
            logging.info("Initializing parameters.")

            inputs = next(self._train_input)

            init_net = jax.pmap(
                lambda *a: self._transform.init(*a, is_training=True),
                axis_name="i",
            )

            # Init uses the same RNG key on all hosts+devices to ensure everyone
            # computes the same initial state.
            init_rng = utils.bcast_local_devices(self.init_rng)
            self._params, self._state = init_net(init_rng, inputs)

        self._lr_schedule = exputils.get_lr_schedule(
            total_steps,
            self.config.optimizer,
        )

        self._optimizer = exputils.make_optimizer(
            self.config.optimizer,
            self._lr_schedule,
        )

        init_opt = jax.pmap(self._optimizer.init, axis_name="i")

        if self._opt_state is None:
            self._opt_state = init_opt(self._params)

    def _build_train_input(
        self,
    ) -> Iterable[Mapping[str, Mapping[str, np.ndarray]]]:
        """Builds the training input.

        For each dataset specified in the config, this will call the appropriate
        constructor for the dataset and then yield elements from each dataset.
        Currently supported datasets are 'selfsup_kinetics_tracks' and 'kubric'.

        Yields:
          Elements of a joint dataset, where each item is a dict.  The keys are
          the dataset names, and the associated values are items generated by
          those dataset classes.
        """

        # Note: a dataset constructor is a function which takes only kwargs from
        # the experiment config, and returns a dataset.
        #
        # A dataset consists of a generator function (calling next() on it will
        # produce a data value)
        dataset_constructors = {
            "kubric": dataset.create_point_tracking_dataset,
        }
        dataset_generators = {}
        for dset_name in self.config.datasets.dataset_names:
            ds_generator = self.create_dataset_generator(
                dataset_constructors,
                dset_name,
                color_augmentation=True,
            )

            dataset_generators[dset_name] = ds_generator

        while True:
            combined_dset = {}
            for dset_name in self.config.datasets.dataset_names:
                next_data = next(dataset_generators[dset_name])
                combined_dset[dset_name] = next_data
            yield combined_dset

    def create_dataset_generator(
        self,
        dataset_constructors: Mapping[str, Callable[..., tf.data.Dataset]],
        dset_name: str,
        color_augmentation: bool = False,
    ) -> Iterator[Mapping[str, np.ndarray]]:
        # Batch data on available devices.
        # Number of devices is unknown when an interpreter reads a config file.
        # Here we re-define batch dims to guide jax for right batching.
        batch_dims = [
            jax.local_device_count(),
            self.config.datasets[dset_name + "_kwargs"].batch_dims,
        ]
        dset_kwargs = dict(self.config.datasets[dset_name + "_kwargs"])
        dset_kwargs["batch_dims"] = []
        ds = dataset_constructors[dset_name](**dset_kwargs)
        if color_augmentation:
            ds = exputils.add_default_data_augmentation(ds)

        for dim in batch_dims[::-1]:
            ds = ds.batch(dim)
        np_ds = tfds.as_numpy(ds)
        return iter(np_ds)

    def _update_func(
        self,
        params: chex.ArrayTree,
        state: chex.ArrayTree,
        opt_state: chex.ArrayTree,
        inputs: chex.ArrayTree,
        rng: chex.PRNGKey,
        global_step: chex.Array,
    ) -> Tuple[chex.Array, chex.Array, chex.Array, Mapping[str, chex.Numeric]]:
        """Applies an update to parameters and returns new state."""

        updates = None
        grads, state, scalars = self.point_prediction.get_gradients(
            params,
            state,
            inputs,
            rng,
            global_step,
            self._transform.apply,
            is_training=True,
        )
        scalars = {**scalars}  # Mutable copy.
        if grads is not None:
            grads = jax.lax.psum(grads, axis_name="i")
            task_updates, opt_state = self._optimizer.update(
                grads,
                opt_state,
                params,
            )

            # We accumulate task_updates across tasks; if this is the first task
            # processed, then updates is None and we just initialize with the
            # updates from the first task.
            if updates is None:
                updates = task_updates
            else:
                updates = jax.tree_map(jnp.add, updates, task_updates)
            scalars["gradient_norm"] = optax.global_norm(grads)

        # Grab the learning rate to log before performing the step.
        learning_rate = self._lr_schedule(global_step)

        # Make the BYOL predictor train faster than everything else.
        def multiply_lr(variable_name, variable_update):
            need_fast_lr = False
            for substr in self.config.fast_variables:
                need_fast_lr = need_fast_lr or any([substr in x for x in variable_name])
            if need_fast_lr:
                logging.info("Boosting learning rate for: %s", "/".join(variable_name))
                return variable_update * 10.0
            return variable_update

        mapping_type = type(updates)

        def map_fn(nested_dict, prefix=None):
            prefix = prefix or []
            result = {}
            for k, v in nested_dict.items():
                if isinstance(v, mapping_type):
                    result[k] = map_fn(v, prefix + [k])
                else:
                    result[k] = multiply_lr(prefix + [k], v)

            return mapping_type(**result)

        updates = map_fn(updates)

        params = optax.apply_updates(params, updates)

        n_params = 0
        for k in params.keys():  # pytype: disable=attribute-error  # numpy-scalars
            for l in params[k]:
                n_params = n_params + np.prod(
                    params[k][l].shape
                )  # pytype: disable=attribute-error  # numpy-scalars

        # Scalars to log (note: we log the mean across all hosts/devices).
        scalars.update(
            {
                "learning_rate": learning_rate,
                "n_params (M)": float(n_params / 1e6),
            }
        )
        scalars = jax.lax.pmean(scalars, axis_name="i")

        return params, state, opt_state, scalars

    #                  _
    #   _____   ____ _| |
    #  / _ \ \ / / _` | |
    # |  __/\ V / (_| | |
    #  \___| \_/ \__,_|_|
    #
    def evaluate(  # pytype: disable=signature-mismatch  # numpy-scalars
        self,
        global_step: chex.Array,
        rng: chex.PRNGKey,
        mode: Optional[str] = None,
        **unused_args,
    ) -> Dict[str, chex.Array]:
        mode = mode or self.mode
        point_prediction_task = self.point_prediction
        forward_fn = self._transform.apply
        eval_scalars = point_prediction_task.evaluate(
            global_step=global_step,
            params=self._params,
            state=self._state,
            rng=rng,
            wrapped_forward_fn=forward_fn,
            mode=mode,
        )
        return {f"eval/{mode}/{key}": value for key, value in eval_scalars.items()}


def main(_):
    flags.mark_flag_as_required("config")
    # Keep TF off the GPU; otherwise it hogs all the memory and leaves none for
    # JAX.
    tf.config.experimental.set_visible_devices([], "GPU")
    tf.config.experimental.set_visible_devices([], "TPU")
    platform.main(
        Experiment,
        sys.argv[1:],
        checkpointer_factory=exputils.NumpyFileCheckpointer.create,
    )


if __name__ == "__main__":
    app.run(main)



"""Main class to perform inference using RTMDet and RTMPose (ONNX)"""

import cv2
import csv
from pathlib import Path
import numpy as np
import supervision as sv

from typing import List, Union, Generator, Literal

import torch

from juxtapose.data import load_inference_source
from juxtapose.detectors import get_detector
from juxtapose.pose.rtmpose import RTMPose
from juxtapose.trackers import Tracker, TRACKER_MAP
from juxtapose.types import (
    DETECTOR_TYPES,
    POSE_ESTIMATOR_TYPES,
    TRACKER_TYPES,
    DEVICE_TYPES,
)

from juxtapose.utils.polygon import PolygonZone

from juxtapose.utils.core import Detections
from juxtapose.utils.plotting import Annotator
from juxtapose.utils.roi import select_roi
from juxtapose.utils.checks import check_imshow
from juxtapose.utils.torch_utils import smart_inference_mode
from juxtapose.utils.ops import xyxy2xyxyxyxy
from juxtapose.utils import (
    LOGGER,
    MACOS,
    WINDOWS,
    colorstr,
    ops,
    get_time,
    DEFAULT_COLOR_PALETTE,
)

from dataclasses import dataclass


@dataclass
class Result:
    im: np.ndarray or None  # shape -> (h, w, c)
    persons: List
    kpts: List  # shape -> (number of humans, 17, 2)
    bboxes: List  # shape -> (number of humans, 4)
    speed: dict  # {'bboxes': ... ms, 'kpts': ... ms} -> used to record the milliseconds of the inference time

    save_dirs: str  # save directory
    name: str  # file name


class RTM:
    """"""

    def __init__(
        self,
        det: DETECTOR_TYPES = "rtmdet-m",
        pose: POSE_ESTIMATOR_TYPES = "rtmpose-m",
        tracker: TRACKER_TYPES = "bytetrack",
        device: DEVICE_TYPES = "cuda" if torch.cuda.is_available() else "cpu",
        annotator=Annotator(),
        captions="person .",
    ) -> None:
        if device == "cuda" and not torch.cuda.is_available():
            LOGGER.info(f"Auto switch to CPU, as you are running without CUDA")
            device = "cpu"

        self.det = self.setup_detector(det, device, captions)
        self.rtmpose = RTMPose(pose.split("-")[1], device=device)
        self.annotator = annotator
        self.box_annotator = sv.BoxAnnotator()

        self.tracker_type = tracker

        if tracker not in TRACKER_MAP.keys():
            self.tracker_type = None

        self.dataset = None
        self.vid_path, self.vid_writer = None, None

    def setup_source(self, source, imgsz=640, vid_stride=1) -> None:
        self.dataset = load_inference_source(
            source=source, imgsz=imgsz, vid_stride=vid_stride
        )
        self.source_type = self.dataset.source_type
        self.vid_path, self.vid_writer = [None] * self.dataset.bs, [
            None
        ] * self.dataset.bs

    def setup_tracker(self) -> None:
        self.tracker = Tracker(self.tracker_type).tracker

    def setup_detector(self, det, device, captions):
        return get_detector(det, device=device, captions=captions)

    def save_preds(self, im0, vid_cap, idx, save_path) -> None:
        """Save video predictions as mp4 at specified path."""
        # Save imgs
        if self.dataset.mode == "image":
            cv2.imwrite(save_path, im0)
        else:  # 'video' or 'stream'
            if self.vid_path[idx] != save_path:  # new video
                self.vid_path[idx] = save_path
                if isinstance(self.vid_writer[idx], cv2.VideoWriter):
                    self.vid_writer[idx].release()  # release previous video writer
                if vid_cap:  # video
                    fps = int(
                        vid_cap.get(cv2.CAP_PROP_FPS)
                    )  # integer required, floats produce error in MP4 codec
                    w = int(vid_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                    h = int(vid_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                else:  # stream
                    fps, w, h = 30, im0.shape[1], im0.shape[0]
                suffix = ".mp4" if MACOS else ".avi" if WINDOWS else ".avi"
                fourcc = "avc1" if MACOS else "WMV2" if WINDOWS else "MJPG"
                save_path = str(Path(save_path).with_suffix(suffix))
                self.vid_writer[idx] = cv2.VideoWriter(
                    save_path, cv2.VideoWriter_fourcc(*fourcc), fps, (w, h)
                )
            self.vid_writer[idx].write(im0)

    def save_csv(self, path, data):
        with open(path, "a", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerow(data)

    def get_labels(self, detections: Detections):
        mp = {0: "person"}
        return np.array(
            [
                f"{mp[label]} {id} {score:.2f}"
                for score, label, id in zip(
                    detections.confidence, detections.labels, detections.track_id
                )
            ]
        )

    def setup_zones(self, w, h, zones: List):
        print(zones)
        zones = [
            (
                PolygonZone(
                    polygon=np.array(zone),
                    frame_resolution_wh=(w, h),
                    triggering_position=sv.Position.CENTER,
                ),
                sv.Color(0, 0, 0).from_hex(DEFAULT_COLOR_PALETTE[idx]),
            )
            for idx, zone in enumerate(zones)
        ]
        return zones

    def setup_roi(self, im, win: str = "roi", type: Literal["rect"] = "rect") -> List:
        w, h, _ = im.shape
        roi_zones = select_roi(im, win, type=type)
        roi_zones = xyxy2xyxyxyxy(roi_zones)
        return self.setup_zones(w, h, roi_zones)

    @smart_inference_mode
    def stream_inference(
        self,
        source,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        zones: List = [],
        # panels
        show=True,
        plot=True,
        plot_bboxes=True,
        save=False,
        save_dirs="",
        verbose=True,
        return_im=False,
    ) -> Result:
        if show:
            check_imshow(warn=True)

        if save:
            save_dirs = Path(get_time() if not save_dirs else save_dirs)
            save_dirs.mkdir(exist_ok=True)

        profilers = (ops.Profile(), ops.Profile(), ops.Profile())  # count the time
        self.setup_source(source)

        if self.tracker_type:
            self.setup_tracker()

        current_source, index = None, 0

        for _, batch in enumerate(self.dataset):
            path, im0s, vid_cap, s = batch
            is_image = s.startswith("image")

            p, im, im0 = Path(path[0]), im0s[0], im0s[0].copy()

            index += 1

            # reset tracker when source changed
            if current_source is None:
                current_source = p
                if len(zones) > 0:
                    w, h, _ = im.shape
                    zones = self.setup_zones(w, h, zones)
                if roi and not zones:
                    zones = self.setup_roi(im)
            elif current_source != p:
                if len(zones) > 0:
                    w, h, _ = im.shape
                    zones = self.setup_zones(w, h, zones)
                if roi and not zones:
                    zones = self.setup_roi(im)
                if self.tracker_type:
                    self.setup_tracker()
                current_source = p
                index = 1

            # multi object detection (detect only person)
            with profilers[0]:
                detections: Detections = self.det(im)

            # multi object tracking (adjust bounding boxes)
            with profilers[1]:
                if self.tracker_type:
                    detections: Detections = self.tracker.update(
                        bboxes=detections.xyxy,
                        confidence=detections.confidence,
                        labels=detections.labels,
                    )
                    track_id = detections.track_id
                else:
                    track_id = np.array([""] * len(detections.xyxy))
                    detections.track_id = track_id

            # filter bboxes based on roi
            if zones and detections:
                masks = set()
                for zone, color in zones:
                    mask = zone.trigger(detections=detections)
                    masks.update(np.where(mask)[0].tolist())
                masks = np.array(
                    [True if i in masks else False for i in range(len(detections.xyxy))]
                )

            # pose estimation (detect 17 keypoints based on the bounding boxes)
            with profilers[2]:
                if detections:
                    # kpts = self.rtmpose(im, bboxes=detections.xyxy)
                    kpts, kpts_scores = self.rtmpose(im, bboxes=detections.xyxy)

            if plot:
                labels = self.get_labels(detections)

                if zones:
                    for zone, color in zones:
                        im = sv.draw_polygon(im, zone.polygon, color)

                    xyxy, labels, track_id, kpts = (
                        detections.xyxy[masks],
                        labels[masks],
                        track_id[masks],
                        kpts[masks],
                    )
                    detections = Detections(xyxy=xyxy, labels=labels, track_id=track_id)
                    if plot_bboxes:
                        self.box_annotator.annotate(im, detections, labels=labels)
                    if detections:
                        self.annotator.draw_kpts(im, kpts)
                        self.annotator.draw_skeletons(im, kpts)
                else:
                    if plot_bboxes:
                        self.box_annotator.annotate(im, detections, labels=labels)
                    if detections:
                        self.annotator.draw_kpts(im, kpts)
                        self.annotator.draw_skeletons(im, kpts)

            if detections:
                result = Result(
                    im=im if return_im else None,
                    kpts=kpts.tolist(),
                    bboxes=detections.xyxy.tolist(),  # detections.xyxy,
                    persons=[
                        {"id": str(i), "kpts": kpt.tolist(), "bboxes": bboxes}
                        for i, kpt, bboxes in zip(
                            detections.track_id, kpts, detections.xyxy.tolist()
                        )
                    ],
                    speed={
                        "bboxes": profilers[0].dt * 1e3 / 1,
                        "track": profilers[1].dt * 1e3 / 1,
                        "kpts": profilers[2].dt * 1e3 / 1,
                    },
                    save_dirs=str(save_dirs) if save else "",
                    name=p.name,
                )
            else:
                result = Result(
                    im=im if return_im else None,
                    kpts=[],
                    bboxes=[],  # detections.xyxy,
                    persons=[],
                    speed={
                        "bboxes": profilers[0].dt * 1e3 / 1,
                        "track": profilers[1].dt * 1e3 / 1,
                        "kpts": profilers[2].dt * 1e3 / 1,
                    },
                    save_dirs=str(save_dirs) if save else "",
                    name=p.name,
                )

            yield result

            if show:
                if is_image:
                    cv2.imshow(result.name, result.im if return_im else im)
                    key = cv2.waitKey(0) & 0xFF
                    if key == ord("q"):
                        cv2.destroyWindow(result.name)
                        continue
                else:
                    cv2.imshow(result.name, result.im if return_im else im)
                    key = cv2.waitKey(1) & 0xFF
                    if key == ord("q"):
                        break
                    elif key == 32:
                        zones = self.setup_roi(im0)

            if verbose:
                LOGGER.info(
                    f"{s}Detected {len(detections.xyxy)} person(s) ({colorstr('green', f'{profilers[0].dt * 1E3:.1f}ms')} | {colorstr('green', f'{profilers[1].dt * 1E3:.1f}ms')} | {colorstr('green', f'{profilers[2].dt * 1E3:.1f}ms')})"
                )

            if save:
                self.save_preds(im, vid_cap, 0, str(save_dirs / p.name))
                self.save_csv(
                    str(save_dirs / p.with_suffix(".csv").name),
                    [
                        str(index),
                        str(
                            [
                                {"id": str(i), "kpts": kpt.tolist(), "bboxes": bboxes}
                                for i, kpt, bboxes in zip(
                                    detections.track_id, kpts, detections.xyxy.tolist()
                                )
                            ]
                        ),
                    ],
                )

        if isinstance(self.vid_writer[-1], cv2.VideoWriter):
            self.vid_writer[-1].release()

        cv2.destroyAllWindows()

    def __call__(
        self,
        source,
        stream=False,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        zones: List = [],
        # panels
        show=True,
        plot=True,
        plot_bboxes=True,
        save=False,
        save_dirs="",
        verbose=True,
        return_im=False,
    ) -> Union[List[Result], Generator[Result, None, None]]:
        if stream:
            return self.stream_inference(
                source,
                timer=timer,
                poi=poi,
                roi=roi,
                zones=zones,
                show=show,
                plot=plot,
                plot_bboxes=plot_bboxes,
                save=save,
                save_dirs=save_dirs,
                verbose=verbose,
                return_im=return_im,
            )
        else:
            return list(
                self.stream_inference(
                    source,
                    timer=timer,
                    poi=poi,
                    roi=roi,
                    zones=zones,
                    show=show,
                    plot=plot,
                    plot_bboxes=plot_bboxes,
                    save=save,
                    save_dirs=save_dirs,
                    verbose=verbose,
                    return_im=return_im,
                )
            )



import functools
import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
import mediapy as media
from pathlib import Path

import cv2
import csv
import supervision as sv

from typing import List, Union, Generator, Literal

from juxtapose.data import load_inference_source
from juxtapose.trackers.tapnet import tapir_model
from juxtapose.trackers.tapnet.utils import transforms
from juxtapose.trackers.tapnet.utils import viz_utils
from juxtapose.utils.downloads import safe_download
from juxtapose.utils.checks import check_imshow
from juxtapose.utils.plotting import Annotator
from juxtapose.utils import (
    LOGGER,
    MACOS,
    WINDOWS,
    colorstr,
    ops,
    get_time,
)

from dataclasses import dataclass


@dataclass
class Result:
    im: np.ndarray  # shape -> (h, w, c)
    tracks: List  # shape -> (number of humans, 17, 2)
    speed: dict  # {'bboxes': ... ms, 'kpts': ... ms} -> used to record the milliseconds of the inference time

    save_dirs: str  # save directory
    name: str  # file name


class Tapnet:
    def __init__(
        self,
        annotator=Annotator(),
    ):
        download_dir = Path("model")
        checkpoint_path = download_dir / f"causal_tapir_checkpoint.npy"

        if not checkpoint_path.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/causal_tapir_checkpoint.npy",
                file=f"causal_tapir_checkpoint",
                dir=download_dir,
            )

        ckpt_state = np.load(checkpoint_path, allow_pickle=True).item()
        params, state = ckpt_state["params"], ckpt_state["state"]

        online_init = hk.transform_with_state(Tapnet.build_online_model_init)
        online_init_apply = jax.jit(online_init.apply)

        online_predict = hk.transform_with_state(Tapnet.build_online_model_predict)
        online_predict_apply = jax.jit(online_predict.apply)

        rng = jax.random.PRNGKey(42)
        self.online_init_apply = functools.partial(
            online_init_apply, params=params, state=state, rng=rng
        )
        self.online_predict_apply = functools.partial(
            online_predict_apply, params=params, state=state, rng=rng
        )
        self.resize_height, self.resize_width = 256, 256

        self.annotator = annotator
        self.box_annotator = sv.BoxAnnotator()

        self.dataset = None
        self.vid_path, self.vid_writer = None, None

    def setup_source(self, source, imgsz=640, vid_stride=1) -> None:
        self.dataset = load_inference_source(
            source=source, imgsz=imgsz, vid_stride=vid_stride
        )
        self.source_type = self.dataset.source_type
        self.vid_path, self.vid_writer = [None] * self.dataset.bs, [
            None
        ] * self.dataset.bs

    def save_preds(self, im0, vid_cap, idx, save_path) -> None:
        """Save video predictions as mp4 at specified path."""
        # Save imgs
        if self.dataset.mode == "image":
            cv2.imwrite(save_path, im0)
        else:  # 'video' or 'stream'
            if self.vid_path[idx] != save_path:  # new video
                self.vid_path[idx] = save_path
                if isinstance(self.vid_writer[idx], cv2.VideoWriter):
                    self.vid_writer[idx].release()  # release previous video writer
                if vid_cap:  # video
                    fps = int(
                        vid_cap.get(cv2.CAP_PROP_FPS)
                    )  # integer required, floats produce error in MP4 codec
                    w = int(vid_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                    h = int(vid_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                else:  # stream
                    fps, w, h = 30, im0.shape[1], im0.shape[0]
                suffix = ".mp4" if MACOS else ".avi" if WINDOWS else ".avi"
                fourcc = "avc1" if MACOS else "WMV2" if WINDOWS else "MJPG"
                save_path = str(Path(save_path).with_suffix(suffix))
                self.vid_writer[idx] = cv2.VideoWriter(
                    save_path, cv2.VideoWriter_fourcc(*fourcc), fps, (w, h)
                )
            self.vid_writer[idx].write(im0)

    def save_csv(self, path, data):
        with open(path, "a", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerow(data)

    def setup_points(self, im):
        fig, ax = plt.subplots(figsize=(10, 5))
        ax.imshow(cv2.cvtColor(im, cv2.COLOR_BGR2RGB))
        ax.axis("off")
        ax.set_title(
            "You can select more than 1 points. After select enough points, run the next cell."
        )

        select_points = []

        # Event handler for mouse clicks
        def on_click(event):
            if event.button == 1 and event.inaxes == ax:  # Left mouse button clicked
                x, y = int(np.round(event.xdata)), int(np.round(event.ydata))

                select_points.append(np.array([x, y]))

                color = (255, 0, 0)
                color = tuple(np.array(color) / 255.0)
                ax.plot(x, y, "o", color=color, markersize=5)
                plt.draw()

        # def close_figure(event):
        #     if event.key == "enter":
        #         plt.close(event.canvas.figure)
        # plt.gcf().canvas.mpl_connect("key_press_event", close_figure)

        fig.canvas.mpl_connect("button_press_event", on_click)
        plt.show()

        return select_points

    def get_query_features(self, im, resized_im: np.ndarray, select_points):
        height, width = im.shape[0], im.shape[1]
        query_points = Tapnet.convert_select_points_to_query_points(0, select_points)
        query_points = transforms.convert_grid_coordinates(
            query_points,
            (1, height, width),
            (1, self.resize_height, self.resize_width),
            coordinate_format="tyx",
        )
        query_features, _ = self.online_init_apply(
            frames=Tapnet.preprocess_frames(resized_im),
            query_points=query_points[None],
        )
        causal_state = Tapnet.construct_initial_causal_state(
            query_points.shape[0], len(query_features.resolutions) - 1
        )
        return query_features, causal_state

    def stream_inference(
        self,
        source,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        # panels
        show=True,
        plot=True,
        save=False,
        save_dirs="",
        verbose=True,
    ) -> Result:
        if show:
            check_imshow(warn=True)

        if save:
            save_dirs = Path(get_time() if not save_dirs else save_dirs)
            save_dirs.mkdir(exist_ok=True)

        profilers = (ops.Profile(), ops.Profile(), ops.Profile())  # count the time
        self.setup_source(source)

        current_source, index = None, 0

        for _, batch in enumerate(self.dataset):
            path, im0s, vid_cap, s = batch
            is_image = s.startswith("image")

            p, im, im0 = Path(path[0]), im0s[0], im0s[0].copy()

            height, width = im.shape[0], im.shape[1]
            resized_im = media.resize_image(im, (self.resize_height, self.resize_width))
            resized_im = np.expand_dims(resized_im, axis=[0, 1])

            index += 1

            # reset tracker when source changed
            if current_source is None:
                current_source = p
                select_points = self.setup_points(im)
                query_features, causal_state = self.get_query_features(
                    im, resized_im, select_points
                )
            elif current_source != p:
                select_points = self.setup_points(im)
                query_features, causal_state = self.get_query_features(
                    im, resized_im, select_points
                )
                current_source = p
                index = 1

            # multi object detection (detect only person)
            with profilers[0]:
                (prediction, causal_state), _ = self.online_predict_apply(
                    frames=resized_im,
                    query_features=query_features,
                    causal_context=causal_state,
                )
                tracks = prediction["tracks"][0]
                occlusions = prediction["occlusion"][0]
                expected_dist = prediction["expected_dist"][0]
                visibles = Tapnet.postprocess_occlusions(occlusions, expected_dist)
                tracks = transforms.convert_grid_coordinates(
                    tracks, (self.resize_width, self.resize_height), (width, height)
                )

            if plot:
                for i, t in enumerate(tracks):
                    if visibles[i]:
                        x, y = t[0][0], t[0][1]
                        cv2.circle(im, (int(x), int(y)), 8, (0, 0, 255), -1)

            result = Result(
                im=im,
                tracks=tracks,  # detections.xyxy,
                speed={
                    "tracks": profilers[0].dt * 1e3 / 1,
                },
                save_dirs=str(save_dirs) if save else "",
                name=p.name,
            )

            yield result

            if show:
                if is_image:
                    cv2.imshow(result.name, result.im)
                    key = cv2.waitKey(0) & 0xFF
                    if key == ord("q"):
                        cv2.destroyWindow(result.name)
                        continue
                else:
                    cv2.imshow(result.name, result.im)
                    key = cv2.waitKey(1) & 0xFF
                    if key == ord("q"):
                        break
                    elif key == 32:
                        pass
                        # zones = self.setup_roi(im0)

            if verbose:
                LOGGER.info(
                    f"{s}Tracking {len(tracks)} point(s) ({colorstr('green', f'{profilers[0].dt * 1E3:.1f}ms')})"
                )

            if save:
                self.save_preds(im, vid_cap, 0, str(save_dirs / p.name))
                self.save_csv(
                    str(save_dirs / p.with_suffix(".csv").name),
                    [
                        str(index),
                        str(np.array(tracks).tolist())
                        # str([{i: kpt} for i, kpt in zip(detections.track_id, kpts)]),
                    ],
                )

        if isinstance(self.vid_writer[-1], cv2.VideoWriter):
            self.vid_writer[-1].release()

        cv2.destroyAllWindows()

    def __call__(
        self,
        source,
        stream=False,
        # interest point
        timer: List = [],
        poi: Literal["point", "box", "text", ""] = "",
        roi: Literal["rect", ""] = "",
        # panels
        show=True,
        plot=True,
        save=False,
        save_dirs="",
        verbose=True,
    ) -> Union[List[Result], Generator[Result, None, None]]:
        if stream:
            return self.stream_inference(
                source,
                timer=timer,
                poi=poi,
                roi=roi,
                show=show,
                plot=plot,
                save=save,
                save_dirs=save_dirs,
                verbose=verbose,
            )
        else:
            return list(
                self.stream_inference(
                    source,
                    timer=timer,
                    poi=poi,
                    roi=roi,
                    show=show,
                    plot=plot,
                    save=save,
                    save_dirs=save_dirs,
                    verbose=verbose,
                )
            )

    @staticmethod
    def build_online_model_init(frames, query_points):
        """Initialize query features for the query points."""
        model = tapir_model.TAPIR(
            use_causal_conv=True, bilinear_interp_with_depthwise_conv=False
        )

        feature_grids = model.get_feature_grids(frames, is_training=False)
        query_features = model.get_query_features(
            frames,
            is_training=False,
            query_points=query_points,
            feature_grids=feature_grids,
        )
        return query_features

    @staticmethod
    def build_online_model_predict(frames, query_features, causal_context):
        """Compute point tracks and occlusions given frames and query points."""
        model = tapir_model.TAPIR(
            use_causal_conv=True, bilinear_interp_with_depthwise_conv=False
        )
        feature_grids = model.get_feature_grids(frames, is_training=False)
        trajectories = model.estimate_trajectories(
            frames.shape[-3:-1],
            is_training=False,
            feature_grids=feature_grids,
            query_features=query_features,
            query_points_in_video=None,
            query_chunk_size=64,
            causal_context=causal_context,
            get_causal_context=True,
        )
        causal_context = trajectories["causal_context"]
        del trajectories["causal_context"]
        return {k: v[-1] for k, v in trajectories.items()}, causal_context

    @staticmethod
    def preprocess_frames(frames):
        """Preprocess frames to model inputs.

        Args:
        frames: [num_frames, height, width, 3], [0, 255], np.uint8

        Returns:
        frames: [num_frames, height, width, 3], [-1, 1], np.float32
        """
        frames = frames.astype(np.float32)
        frames = frames / 255 * 2 - 1
        return frames

    @staticmethod
    def postprocess_occlusions(occlusions, expected_dist):
        """Postprocess occlusions to boolean visible flag.

        Args:
        occlusions: [num_points, num_frames], [-inf, inf], np.float32

        Returns:
        visibles: [num_points, num_frames], bool
        """
        pred_occ = jax.nn.sigmoid(occlusions)
        pred_occ = 1 - (1 - pred_occ) * (1 - jax.nn.sigmoid(expected_dist))
        visibles = pred_occ < 0.5  # threshold
        return visibles

    @staticmethod
    def sample_random_points(frame_max_idx, height, width, num_points):
        """Sample random points with (time, height, width) order."""
        y = np.random.randint(0, height, (num_points, 1))
        x = np.random.randint(0, width, (num_points, 1))
        t = np.random.randint(0, frame_max_idx + 1, (num_points, 1))
        points = np.concatenate((t, y, x), axis=-1).astype(np.int32)  # [num_points, 3]
        return points

    @staticmethod
    def construct_initial_causal_state(num_points, num_resolutions):
        value_shapes = {
            "tapir/~/pips_mlp_mixer/block_1_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_1_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_2_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_2_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_3_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_3_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_4_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_4_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_5_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_5_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_6_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_6_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_7_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_7_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_8_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_8_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_9_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_9_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_10_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_10_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_11_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_11_causal_2": (1, num_points, 2, 2048),
            "tapir/~/pips_mlp_mixer/block_causal_1": (1, num_points, 2, 512),
            "tapir/~/pips_mlp_mixer/block_causal_2": (1, num_points, 2, 2048),
        }
        fake_ret = {k: jnp.zeros(v, dtype=jnp.float32) for k, v in value_shapes.items()}
        return [fake_ret] * num_resolutions * 4

    @staticmethod
    def convert_select_points_to_query_points(frame, points):
        """Convert select points to query points.

        Args:
        points: [num_points, 2], [t, y, x]
        Returns:
        query_points: [num_points, 3], [t, y, x]
        """
        points = np.stack(points)
        query_points = np.zeros(shape=(points.shape[0], 3), dtype=np.float32)
        query_points[:, 0] = frame
        query_points[:, 1] = points[:, 1]
        query_points[:, 2] = points[:, 0]
        return query_points



import cv2
import supervision as sv
from typing import Literal
from juxtapose.utils import LOGGER


def select_roi(
    im,
    win: str = "roi",
    type: Literal["rect"] = "rect",
    color=sv.ColorPalette.default(),
):
    x0, y0 = -1, -1
    points, bboxes = [], []
    img, img4show = im.copy(), im.copy()
    col = ()

    def BOX(event, x, y, flags, param):
        nonlocal x0, y0, img4show, img, col
        if event == cv2.EVENT_LBUTTONDOWN:
            x0, y0 = x, y
        elif event == cv2.EVENT_MOUSEMOVE and flags == cv2.EVENT_FLAG_LBUTTON:
            img4show = img.copy()
            col = color.by_idx(len(bboxes))
            cv2.rectangle(img4show, (x0, y0), (x, y), col.as_bgr(), 2)
        elif event == cv2.EVENT_LBUTTONUP:
            img = img4show
            bboxes.append([x0, y0, x, y])

    def POINT(event, x, y, flags, param):
        nonlocal x0, y0, img4show, img
        if event == cv2.EVENT_LBUTTONDOWN:
            x0, y0 = x, y
            points.append([x0, y0])
        elif event == cv2.EVENT_LBUTTONUP:
            cv2.circle(img4show, (x0, y0), 4, color.by_idx(len(points)).as_bgr(), 6)
            img = img4show

    cv2.namedWindow(win)
    cv2.setMouseCallback(win, BOX if type == "rect" else POINT)

    # display the window
    while True:
        cv2.putText(
            img4show,
            "SELECT ROI",
            (200, 200),
            cv2.FONT_HERSHEY_DUPLEX,
            2,
            (255, 255, 0),
        )
        cv2.imshow(win, img4show)
        k = cv2.waitKey(1) & 0xFF
        if k == 114 or k == 82:  # RESET with 'r' or 'R'
            img = im.copy()
            img4show = im.copy()
            points = []
            bboxes = []
        elif k == 113 or k == 81 or k == 27:  # return none with 'q' or 'Q' or 'ESC'
            # return [], []
            break
        elif k == 32 or k == 13:  # DONE with 'SPACE' or 'ENTER'
            break
        elif k != 255:
            LOGGER.info(
                """After selected the ROI,
Press SPACE/ENTER     to continue
Press r/R             to restart
Press q/Q/ESC         to quit
                """
            )

    cv2.destroyWindow(win)

    if type == "rect":
        return bboxes
    else:
        return points, [i + 1 for i in range(len(points))]
